import matplotlib.pyplot as plt
import qtawesome

from PyQt5.QtWidgets import (QPushButton, QLabel, QCheckBox, QColorDialog,
                             QComboBox, QSpinBox, QDoubleSpinBox, QApplication)
from PyQt5.QtGui import QIcon

from pyrateshield.gui.item_views import EditValueWidget



class View(EditValueWidget):
    def create_widgets(self):        
        self.grid_size = QSpinBox(self)
        self.grid_size.setRange(1, 9999)

        self.cmap_name = QComboBox(self)
        self.cmap_name.addItems(plt.colormaps())
        
        self.vmin = QDoubleSpinBox(self)
        self.vmin.setSingleStep(0.1)
        self.vmax = QDoubleSpinBox(self)
        self.vmax.setSingleStep(0.1)
        self.alpha = QDoubleSpinBox(self)
        self.alpha.setRange(0, 1)
        self.alpha.setSingleStep(0.1)
        self.alpha_gradient = QCheckBox(self)
        
        self.contour_line_widgets = []
        for i in range(5):
            clw = {
                "active": QCheckBox(self),
                "level": QDoubleSpinBox(self, decimals=3),
                "color_str": "",
                "color_btn": QPushButton(self),
                "dashing": QComboBox(self),
                "thickness": QDoubleSpinBox(self),
            }
            clw["color_btn"].setMinimumWidth(50)
            clw["color_btn"].clicked.connect(lambda state, x=i: self.linestyle_btn_callback(x))
            clw["dashing"].addItems(("solid", "dotted", "dashed", "dashdot"))
            
            self.contour_line_widgets.append(clw)
            
        self.legend = QCheckBox("Show legend")
        
        self.save_button = QPushButton('Apply')
        self.close_button = QPushButton('Close')
    
    def linestyle_btn_callback(self, index):
        color = QColorDialog(self).getColor()
        if color.isValid():
            self.set_linestyle_color(index, color.name())
            
    def set_linestyle_color(self, index, color):
        self.contour_line_widgets[index]["color_btn"].setStyleSheet('background-color: %s' % color)
        self.contour_line_widgets[index]["color_str"] = color
    
    def create_layout(self):            
        self.setWindowTitle('Dosemap style')
        icon = icon = qtawesome.icon('mdi.palette')
        self.setWindowIcon(QIcon(icon))
        
        row = 0
        
        label = QLabel('Dose map resolution')
        label.setStyleSheet('font-weight: bold')
        self.layout.addWidget(label, row, 0)
        
        row += 1
        
        self.layout.addWidget(QLabel('Matrix size:'), row, 0)
        self.layout.addWidget(self.grid_size, row, 1)
        
        row += 1
        
        text = "The dose map is calculated on an NxM grid, superimposed on "\
               "the current view of the project. This parameter is used to "\
               "specify the number of columns, which automatically fixes the "\
               "number of rows through the aspect ratio of the view. A higher "\
               "value (>200) results in a better looking dose map but may "\
               "cause longer computtion times, most notable for large projects."
        
        label = QLabel(text)
        label.setWordWrap(True)
        self.layout.addWidget(label, row, 0, 1, 4)        
        
        row += 1
        self.interpolate = QCheckBox('Interpolate dosemap to floorplan (slower)')
        self.layout.addWidget(self.interpolate, row, 0, 1, 4)
        
        row += 1
        self.layout.addWidget(QLabel(''), row, 0)
        
        row += 1
                
        label = QLabel('Heat map')
        label.setStyleSheet('font-weight: bold')
        self.layout.addWidget(label, row, 0)
        
        row += 1
        
        self.layout.addWidget(QLabel('Color map:'), row, 0)
        self.layout.addWidget(self.cmap_name, row, 1)
        
        row += 1
        
        self.layout.addWidget(QLabel('Vmin:'), row, 0)
        self.layout.addWidget(self.vmin, row, 1)
        
        row += 1
        
        self.layout.addWidget(QLabel('Vmax:'), row, 0)
        self.layout.addWidget(self.vmax, row, 1)
        
        row += 1
        
        self.layout.addWidget(QLabel('Alpha:'), row, 0)
        self.layout.addWidget(self.alpha, row, 1)
        
        row += 1
        
        self.layout.addWidget(QLabel('Alpha gradient:'), row, 0)
        self.layout.addWidget(self.alpha_gradient, row, 1)
                        
        row += 1
        self.layout.addWidget(QLabel(''), row, 0)
        
        row += 1
        label = QLabel('Contour lines')
        label.setStyleSheet('font-weight: bold')
        self.layout.addWidget(label, row, 0)
        
        row += 1
        self.layout.addWidget(QLabel('Level'), row, 0)
        self.layout.addWidget(QLabel('Style'), row, 1)
        self.layout.addWidget(QLabel('Color'), row, 2)
        self.layout.addWidget(QLabel('Thickness'), row, 3)
        self.layout.addWidget(QLabel('Show'), row, 4)
        
        for i, clw in enumerate(self.contour_line_widgets):
            row += 1
            
            self.layout.addWidget(clw["level"], row, 0)
            self.layout.addWidget(clw["dashing"], row, 1)
            self.layout.addWidget(clw["color_btn"], row, 2)
            self.layout.addWidget(clw["thickness"], row, 3)
            self.layout.addWidget(clw["active"], row, 4)
        
        row += 1
        
        self.layout.addWidget(self.legend)
        
        row += 1
        
        self.layout.addWidget(self.save_button, row, 0)
        self.layout.addWidget(self.close_button, row, 1)

        

class Controller:

    def __init__(self, model, parent=None):
        self.model = model        
        self.view = View()
        self.parent = parent
        
        self.view.grid_size.setValue(self.model.dosemap.grid_matrix_size)        
                
        value = self.view.cmap_name.findText(self.model.dosemap_style.cmap_name)
        self.view.cmap_name.setCurrentIndex(value)
        
        self.view.vmin.setValue(self.model.dosemap_style.vmin)
        self.view.vmax.setValue(self.model.dosemap_style.vmax)
        self.view.alpha.setValue(self.model.dosemap_style.alpha)
        self.view.alpha_gradient.setChecked(self.model.dosemap_style.alpha_gradient)
        self.view.interpolate.setChecked(self.model.dosemap_style.interpolate)
        
        for i, clw in enumerate(self.view.contour_line_widgets):
            if i < len(self.model.dosemap_style.contour_lines):
                cl = self.model.dosemap_style.contour_lines[i]
                level, color, dashing, thickness, active = cl
            else:
                level, color, dashing, thickness, active = (1.0, "black", "solid", 1.5, False)
                    
            clw["level"].setValue(level)
            self.view.set_linestyle_color(i, color)
            clw["dashing"].setCurrentIndex( clw["dashing"].findText(dashing) )
            clw["thickness"].setValue(thickness)
            clw["active"].setChecked(active)
        
        self.view.legend.setChecked(self.model.dosemap_style.show_legend)
        
        
        self.view.save_button.clicked.connect(self.apply)
        self.view.close_button.clicked.connect(self.view.close)
        
        
    
    def apply(self):
        cur_grid_size = self.model.dosemap.grid_matrix_size
        self.model.dosemap.grid_matrix_size = self.view.grid_size.value()
        self.model.dosemap_style.cmap_name = self.view.cmap_name.currentText()
        self.model.dosemap_style.vmin = self.view.vmin.value()
        self.model.dosemap_style.vmax = self.view.vmax.value()
        self.model.dosemap_style.alpha = self.view.alpha.value()
        self.model.dosemap_style.alpha_gradient = bool(self.view.alpha_gradient.checkState())
        self.model.dosemap_style.show_legend = self.view.legend.isChecked()
        self.model.dosemap_style.interpolate = self.view.interpolate.isChecked()
        contour_lines = []
        for i, clw in enumerate(self.view.contour_line_widgets):
            contour_lines.append([
                    clw["level"].value(),
                    clw["color_str"],
                    clw["dashing"].currentText(),
                    clw["thickness"].value(),
                    bool(clw["active"].checkState()),
                ])        
        self.model.dosemap_style.contour_lines = contour_lines
        self.parent.graphics().viewport().update()
        

if __name__ == "__main__":
    app = QApplication([])
    from pyrateshield.model import Model
    project = '../../example_projects/SmallProject/project.zip'

    model = Model.load_from_project_file(project)
    controller = Controller(model)
    window = controller.view
    window.show()    
    app.exec_()
    
    
    
    



# -*- coding: utf-8 -*-
"""
Created on Wed Aug  2 11:54:13 2023

@author: r757021
"""


from PyQt5.QtWidgets import (QApplication, QGraphicsScene,  
                             QGraphicsRectItem, QGraphicsTextItem, 
                             QGraphicsLineItem)
                             
from PyQt5.QtGui import QPen, QColor, QBrush, QFont
from PyQt5.QtCore import Qt


class Legend(QGraphicsScene):
    
    FONT_SIZE       = 10
    

    def __init__(self, model=None):
        self.model = model
       
        super().__init__()

        self.drawItems()
    
    def dpiX(self):
        return QApplication.desktop().physicalDpiX()
    
    def dpiY(self):
        return QApplication.desktop().physicalDpiY()
    
    def spacingX(self):
        return 0.1 * self.dpiX()
    
    def spacingY(self):
        return 0.1 * self.dpiY()
    
    def marginX(self):
        return 0.1 * self.dpiX()
    
    def marginY(self):
        return 0.1 * self.dpiY()
    
    def lineLength(self):
        return 0.5 * self.dpiX()

    def left(self):
        return 0.2 * self.dpiX()
    
    def top(self):
        return 0.2 * self.dpiY()
            
    def font(self):
        font = QFont()
        font.setPointSizeF(self.FONT_SIZE)
        return font

    def drawItems(self):
        self.clear()
        self.drawLines()
        self.addItem(self._rect())
        self.setSceneRect(self.itemsBoundingRect())

    def _rect(self):
        bounds = self.itemsBoundingRect()
      
        rect = QGraphicsRectItem(bounds.x() - self.marginX(),
                                 bounds.y() - self.marginY(),
                                 bounds.width() + 2 * self.marginX(),
                                 bounds.height() + 2 * self.marginY())
                                        
        rect.setZValue(30)
        brush = QBrush()
        brush.setColor(QColor(255, 255, 255, 200))
        brush.setStyle(Qt.SolidPattern)
        rect.setBrush(brush)
        return rect
    
    def _linePen(self, style, color, width):
        style = self.model.get_qt_linestyle(style)
        color = self.model.get_qt_color(color)
        
        pen = QPen()
        pen.setColor(color)
        pen.setStyle(style)
        pen.setWidthF(width*self.dpiX()/72)
        return pen

    
    def entry(self, msv, color, style, linewidth, row):                               
        ll      = self.lineLength()
        sx, sy  = self.spacingX(), self.spacingY()
        
        text = QGraphicsTextItem(str(msv) + ' mSv')
        text.setFont(self.font())
        th = text.boundingRect().height()
        
        text.setPos(ll + sx, row * th + (row-1) * sy)
        
        ly = text.pos().y() + 0.5 * th  
        line = QGraphicsLineItem(0, ly, ll, ly)
        line.setPen(self._linePen(style, color, linewidth))
     
        text.setZValue(31)
        line.setZValue(31)
        
        return line, text
        
    
    def drawLines(self):    
        lines = [line for line in self.model.contour_lines\
                 if line[4]]
        for i, line in enumerate(lines):
            msv, color, style, linewidth, enabled = line
            if not enabled: continue
            line, text = self.entry(msv, color, style, linewidth, i)
        
            self.addItem(text)
            self.addItem(line)
            
    
    



import qtawesome                             
import pandas as pd

from PyQt5.QtWidgets import QLabel, QTableView,  QComboBox, QApplication                            
from PyQt5.QtGui import QIcon


from pyrateshield.constants import CONSTANTS
from pyrateshield.labels import ENERGY_KEV, ABUNDANCE, PARENT, ISOTOPES
from pyrateshield.gui.item_views import EditValueWidget
from pyrateshield.gui.critical_point_controller import PandasModel

INFO_LABEL =('Isotopes can not be added by the user at the moment. '
             'To request a new isotope send an e-mail to the developers '
             ' of pyrateshield.'
             '\n'
             'For some isotopes daughters are included. For daughter nuclei '
             'the abundance is corrected for the branching ratio.'
             '\n'
             'Pyshield neglects gamma rays with energies smaller than 30 keV '
             'because no buildup data is available for these energies.')


class View(EditValueWidget):
    explanation = INFO_LABEL
    def create_widgets(self):
        self.isotope_combo = QComboBox()
        self.isotope_label = QLabel()
        self.table_view = QTableView()
        
    def create_layout(self):
        self.setWindowTitle(ISOTOPES)
        icon = qtawesome.icon('mdi.atom')
        self.setWindowIcon(QIcon(icon))
        self.layout.addWidget(self.isotope_combo, 0, 0)
        self.layout.addWidget(self.table_view, 1, 0)

    def set_stretch(self):
        
        
        
        if hasattr(self, 'explanation'):
            label = QLabel(self.explanation)
            label.setWordWrap(True)
            self.layout.addWidget(label)
            
        self.setLayout(self.layout)
        
    @property
    def isotope(self):
        return self.isotope_combo.currentText()

class Controller:
    def __init__(self, view=None):
        if view is None:
            view = View()
        
        self.view = view
        
        
        self.add_callbacks()
        
        
        self.view.isotope_combo.addItems(self.isotopes)
        
    @property
    def model(self):
        return PandasModel(self.data)
        
    @property
    def data(self):
        data = []
        if self.view.isotope != '':
            isotope = CONSTANTS.get_isotope_by_name(self.view.isotope)
            for parent, keV, abundance in isotope.spectrum_with_parent:
                entry = {PARENT: parent, ENERGY_KEV: keV, ABUNDANCE: abundance}
                data += [entry]
        return pd.DataFrame(data)
        
    def add_callbacks(self):
        self.view.isotope_combo.currentIndexChanged.connect(self.update_isotope)
        
    @property
    def isotopes(self):
        return [item.name for item in CONSTANTS.isotopes]
    
    def update_isotope(self):
        self.view.table_view.setModel(self.model)

        
if __name__ == "__main__":
    app = QApplication([])
    
    controller = Controller()
    window = controller.view
    window.show()    
    app.exec_()


from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QMessageBox
from pyrateshield.gui.main_controller import MainController
from pyrateshield.dosemapper import Dosemapper
from pyrateshield.model import Model
from pyrateshield import labels
from pyrateshield import __pkg_root__, __version__
from PyQt5.QtGui import QIcon

#from pyrateshield.gui.mpl_items_controller import MplItemsPlotter
#from pyrateshield.gui.mpl_controller import plot_dosemap
import matplotlib.pyplot as plt

import os
import multiprocessing
import argparse

_LICENCE_FILE = os.path.join(__pkg_root__, 'LICENSE')

DISCLAIMER =\
("PyrateShield is free to use under the GNU GPLv3 license. The developers do "
 "not take any responsibility for any damages that might arise from using this "
 "software.\n\n"
 "This is a beta version, distributed for the purpose of testing and "
 "validation. Use at your own risk!\n\n"
 "Marcel Segbers (m.segbers@erasmusmc.nl)\n"
 "Rob van Rooij (r.vanrooij-3@umcutrecht.nl)\n\n"
 "Pyrateshield version: " + __version__)


def approve_disclaimer():
    with open (_LICENCE_FILE, "r") as fn:
        LICENCE=''.join(fn.readlines())

    msg = QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Information)
    msg.setText("Disclaimer!")
    msg.setInformativeText(DISCLAIMER)
    msg.setWindowTitle("Disclaimer")
    msg.setDetailedText(LICENCE)
    msg.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
    approvement = msg.exec() == QMessageBox.Ok
    
    return approvement

def get_arguments():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--pyshield', action='store_true')
    group.add_argument('--radtracer', action='store_true')
    parser.add_argument('project', nargs='*')
    args = parser.parse_args()
    
    project_path = args.project[0] if len(args.project) else None
    if project_path is not None and not os.path.exists(project_path):
        raise ValueError(f"Path does not exist: {project_path}")
    
    if args.radtracer:
        engine = labels.RADTRACER
    elif args.pyshield:
        engine = labels.PYSHIELD
    else:
        engine = None
    
    return project_path, engine

def launch_app(dm, project):
    QtWidgets.QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    
    app = QtWidgets.QApplication([])   
    icon = os.path.join(os.path.split(__file__)[0], 'gui', 'icon.png')
    app.setWindowIcon(QIcon(icon))  
    
    if not approve_disclaimer():
        return None
        
    controller = MainController(dosemapper=dm, model=project)
    window = controller.view
    window.showMaximized()
    window.show()
    app.exec_()
    return controller


def stand_alone(dm, project, engine):
    project.dosemap.engine = engine
    dosemap = dm.get_dosemap(project)
    report = dm.get_critical_points(project)    
    print(report)
    
    ax = plt.axes()    
    if project.floorplan.image is not None:
        ax.imshow(project.floorplan.image, extent=project.floorplan.extent)
    MplItemsPlotter(model=project, axes=ax)
    plot_dosemap(ax, project, dosemap)
    plt.show()
    

def main():
    project_path, engine = get_arguments()
    
    if project_path:
        project = Model.load_from_project_file(project_path)
    else:
        project = None
    
    if engine and project is None:
        print("Nothing to execute, specify project file")
        exit()

    controller = None
    with Dosemapper(multi_cpu=True) as dm:
        if engine:
            stand_alone(dm, project, engine)
        else:
            controller = launch_app(dm, project)
        
    return controller

if __name__ == "__main__":
    multiprocessing.freeze_support()
    controller = main()
    




import subprocess
import random
import os
import time
import numpy as np
import itertools
import multiprocessing
from datetime import timedelta

from constants import Isotopes, Materials


NR_OF_PHOTONS = "1E7"   #Keep this a string
REPEATS = 4
BASE_FOLDER = os.path.join("Simulations", NR_OF_PHOTONS)
CPUS_PER_TASK = multiprocessing.cpu_count()//2 

THICKNESS = np.concatenate([
    np.arange( 0,   2, 0.5),  # 4
    np.arange( 2,  10, 1),    # 8
    np.arange(10,  30, 2),    # 10
    np.arange(30, 181, 5),    # 31
])
THICKNESS_LEAD = THICKNESS * 0.1

SKIP_ISOTOPE = ["Ge-68"] # Skip simulation for these isotopes: Ge-68 is identical to Ga-68

isotopes = Isotopes()
materials = Materials()


SIM_TIME = []

def simulate(isotope, self_shielding, material, repeat_idx, todo):
    global SIM_TIME
    
    print(isotope, self_shielding, material)
    
    with open(f"Templates/{self_shielding}.i") as f:
        template_orig = f.read()
        
    folder = os.path.join(f"{BASE_FOLDER}_{repeat_idx+1}", f"{isotope}_{self_shielding}_{material}")
    if not os.path.isdir(folder):
        os.makedirs(folder)
        
    if material in ("Lead", "Tungsten"):
        thickness = THICKNESS_LEAD
    else:
        thickness = THICKNESS
    
    for th in thickness:
        start_time = time.time()
        
        thickness_str = "%.2f"%float(th)
        
        inpf = thickness_str+".i"
        inpf_path = os.path.join(folder, inpf)
        if os.path.exists(inpf_path):
            print(f"Already done {inpf_path}")
            continue

        if os.path.exists("./stop"):
            print("Got stop command...")
            exit()

        open(inpf_path, 'a').close()
        print(f"Run {inpf_path}")

        seed = random.getrandbits(44)
        if not seed % 2:
            seed += 1
            
        template = template_orig.replace('<THICKNESS>', thickness_str)
        template = template.replace('<OUTFILE>', inpf[:-2])
        template = template.replace('<MATERIAL>', materials.get_MDEF(material))
        template = template.replace('<SDEF>', isotopes.get_SDEF(isotope))        
        template = template.replace('<NPS>', NR_OF_PHOTONS)
        template = template.replace('<RAND_SEED>', str(seed))
        
        with open(inpf_path, 'w') as f:
            f.write(template)
        
        n_cpu = max(1, CPUS_PER_TASK)
        subprocess.call(['mcnp6', f'I={inpf}', "TASKS", str(n_cpu)], cwd=folder)    
        
        if os.path.exists(inpf_path[:-2]+".r"):
            os.remove(inpf_path[:-2]+".r")
        
        #### Estimate remaining time ###        
        SIM_TIME.append(time.time()-start_time)
        if len(SIM_TIME) > len(THICKNESS):
            SIM_TIME.pop(0)
        average_time = np.mean(SIM_TIME)
        todo[0] -= 1
        
        print("\nRemaining nr of simulations:", todo[0])
        print("Average time per simulation:", round(average_time,1), "sec")
        print("Estimated time left:", timedelta(seconds=todo[0]*average_time), "\n\n")


def get_todo(isotope_list, material_list, self_shieldings):
    total = 0
    todo = 0
    for repeat_idx in range(REPEATS):        
        for isotope, material, self_shielding in itertools.product(isotope_list, material_list, self_shieldings):
            if isotope in SKIP_ISOTOPE:
                continue
                        
            folder = os.path.join(f"{BASE_FOLDER}_{repeat_idx+1}", f"{isotope}_{self_shielding}_{material}")            
            if material in ("Lead", "Tungsten"):
                thickness = THICKNESS_LEAD
            else:
                thickness = THICKNESS            
            for th in thickness:
                thickness_str = "%.2f"%float(th)                
                inpf = thickness_str+".i"
                inpf_path = os.path.join(folder, inpf)
                
                total += 1
                if not os.path.exists(inpf_path):
                    todo += 1
    print("Total:", total, "Todo:", todo)
    return [todo]

if __name__ == "__main__":        
    isotope_list = isotopes.names()
    material_list = materials.names()
    self_shieldings = ("None", "Body")
    
    #material_list.remove("Tungsten")
    #isotope_list.remove("Rb-82")
    todo = get_todo(isotope_list, material_list, self_shieldings)
    
    for repeat_idx in range(REPEATS):        
        for isotope, material, self_shielding in itertools.product(isotope_list, material_list, self_shieldings):
            if isotope in SKIP_ISOTOPE:
                continue
                     
            simulate(isotope, self_shielding, material, repeat_idx, todo)            
            




import os
import numpy as np
from scipy.optimize import curve_fit
from scipy import interpolate
import matplotlib.pyplot as plt
import pickle
from constants import Isotopes, Materials


PICKLE_FILE = "MCNP.pickle"
BASE_FOLDER = os.path.join("Simulations","1E7")

DUPLICATE_RESULTS = [("Ga-68", "Ge-68"),] # Duplicate results for these isotopes: Ga-68 will also be stored as Ge-68 

mu_rho = np.array([
    (1.00000E-03, 3.606E+03, 3.599E+03),
    (1.50000E-03, 1.191E+03, 1.188E+03),
    (2.00000E-03, 5.279E+02, 5.262E+02),
    (3.00000E-03, 1.625E+02, 1.614E+02),
    (3.20290E-03, 1.340E+02, 1.330E+02),
    (3.20290E-03, 1.485E+02, 1.460E+02),
    (4.00000E-03, 7.788E+01, 7.636E+01),
    (5.00000E-03, 4.027E+01, 3.931E+01),
    (6.00000E-03, 2.341E+01, 2.270E+01),
    (8.00000E-03, 9.921E+00, 9.446E+00),
    (1.00000E-02, 5.120E+00, 4.742E+00),
    (1.50000E-02, 1.614E+00, 1.334E+00),
    (2.00000E-02, 7.779E-01, 5.389E-01),
    (3.00000E-02, 3.538E-01, 1.537E-01),
    (4.00000E-02, 2.485E-01, 6.833E-02),
    (5.00000E-02, 2.080E-01, 4.098E-02),
    (6.00000E-02, 1.875E-01, 3.041E-02),
    (8.00000E-02, 1.662E-01, 2.407E-02),
    (1.00000E-01, 1.541E-01, 2.325E-02),
    (1.50000E-01, 1.356E-01, 2.496E-02),
    (2.00000E-01, 1.233E-01, 2.672E-02),
    (3.00000E-01, 1.067E-01, 2.872E-02),
    (4.00000E-01, 9.549E-02, 2.949E-02),
    (5.00000E-01, 8.712E-02, 2.966E-02),
    (6.00000E-01, 8.055E-02, 2.953E-02),
    (8.00000E-01, 7.074E-02, 2.882E-02),
    (1.00000E+00, 6.358E-02, 2.789E-02),
    (1.25000E+00, 5.687E-02, 2.666E-02),
    (1.50000E+00, 5.175E-02, 2.547E-02),
    (2.00000E+00, 4.447E-02, 2.345E-02),
    (3.00000E+00, 3.581E-02, 2.057E-02),
    (4.00000E+00, 3.079E-02, 1.870E-02),
    (5.00000E+00, 2.751E-02, 1.740E-02),
    (6.00000E+00, 2.522E-02, 1.647E-02),
    (8.00000E+00, 2.225E-02, 1.525E-02),
    (1.00000E+01, 2.045E-02, 1.450E-02),
    (1.50000E+01, 1.810E-02, 1.353E-02),
    (2.00000E+01, 1.705E-02, 1.311E-02),
])
_log_interp = interpolate.interp1d(np.log(mu_rho[:,0]*1e3), np.log(mu_rho[:,2]), fill_value="extrapolate")

def linear_energy_absorption_coeff_air(energy_kev):
    return np.exp(_log_interp(np.log(energy_kev))) / 10 # cm^2/g --> m^2/kg

def ratio_H10_air_kerma(energy_keV):
    # source https://nucleonica.com/wiki/index.php?title=Help:Dose_Rate_Constants%2B%2B
    E0 = 9.85
    xi = np.log(energy_keV / E0)
    r = xi / (1.465 * xi**2 - 4.414 * xi + 4.789) \
        + 0.7006 * np.arctan(0.6519 * xi)
    return np.maximum(r, 0)

def flux_to_h10(energy_kev, flux):
    energy_J = energy_kev * 1.60217662e-16    
    kerma_air_rate = flux * energy_J * linear_energy_absorption_coeff_air(energy_kev)
    return np.dot(ratio_H10_air_kerma(energy_kev), kerma_air_rate)


def getData(folder, phot_per_decay):    
    dose = []

    filenames = [fn for fn in os.listdir(folder) if fn.endswith(".o")]
    for fn in filenames:
        if "conflicted" in fn:
            continue
        
        thickness = float(fn[:-2])        
        path = os.path.join(folder, fn)   
        found = False 
        with open(path) as f:
            energy = []
            flux = []
            for line in f.readlines():
                line = line.strip().split()
                if not len(line): continue
                if line[0] == 'detector' and line[1] == 'located':
                    found = True
                if found:
                    try:
                        energy.append(float(line[0]) * 1e3)
                        flux.append(float(line[1]))
                    except ValueError:
                        if line[0] == 'total':
                            error = float(line[2])
                            break
            
            ### Note: source-detector distance is 400 cm, MCNP outputs flux as MeV/cm^2 (hence the 400**2)
            flux = np.array(flux) * phot_per_decay * 400**2  * 3600 * 1e12 # 3600 s/hr, 1e12: Sv->uSv, Bq->MBq
            
            dose.append([thickness, flux_to_h10(np.array(energy), np.array(flux)), error])
               
    return np.array(sorted(dose))


def fix_spikes(dose, material):
    dx = dose[-1,0]-dose[-2,0]
    
    ### Moving average    
    x = np.arange(dose[-1,0], 0-0.1*dx, -dx/2)[::-1]    
    y = np.interp(x, dose[:,0], np.log(dose[:,1]))

    w = 7
    ma = np.copy(y)
    ma[w//2:-(w//2)] = np.convolve(y, np.ones(w), 'valid') / w
        
    indices = np.where( (dose[1:,0]-dose[:-1,0]) >= 0.5*dx )[0] + 1
    
    newdose = np.copy(dose)
    newdose[indices,1] = np.exp( np.interp(dose[indices,0], x, ma) )
    
    
    # Recursively remove outliers
    remove = np.where(dose[:,1] / newdose[:,1] > 3)[0]
    if len(remove):
        print("Removed:", len(remove))
        plt.plot(dose[remove,0], dose[remove,1], marker='o', ms=10, c='k', fillstyle='none')    
        newdose = np.delete(newdose, remove, axis=0)  
        return fix_spikes(newdose, material)
    
    return newdose



if __name__ == "__main__":        
    try:
        results = pickle.load(open(PICKLE_FILE, "rb"))
    except FileNotFoundError:
        results = {}
    
    isotopes = Isotopes()
    materials = Materials()
    
    for material in materials.names():
        if material == "Tungsten": continue
        for isotope in isotopes.names():
            if isotope == "Rb-82": continue
        
            spectrum = np.array(isotopes.get_spectrum(isotope))
            phot_per_decay = spectrum[:,1].sum()/100
            print(f"\n{isotope} yield:", "%.3f"%phot_per_decay)
                
            color = next(plt.gca()._get_lines.prop_cycler)['color']
            for self_shielding in ("None", "Body"):
                
                all_doses = []
                for base_folder in [BASE_FOLDER+f"_{i}" for i in range(10)]:
                    folder = os.path.join(base_folder, f"{isotope}_{self_shielding}_{material}")
                    
                    if os.path.exists(folder):
                        all_doses.append( getData(folder, phot_per_decay) )

                if not len(all_doses):
                    continue
                
                # Get the values with the lowest statistical uncertainty
                all_doses = np.array(all_doses)
                print(all_doses.shape)
                min_arr = np.argmin(all_doses, axis=0)[:,2]                
                dose = np.empty_like(all_doses[0])                                
                for j in range(len(dose)):
                    dose[j] = all_doses[min_arr[j]][j]
                                
                h_10 = dose[0,1]
                print("h(10): %.5f"%h_10, self_shielding)
                
                dose[:,1] /= h_10                
                dose = dose[np.where(dose[:,1])[0]] # Exclude dose values of 0                
                
                ls = '-' if self_shielding == "None" else '--'
                ms = '.' if self_shielding == "None" else '+'
                
                plt.plot(dose[:,0], dose[:,1], ms, color=color, label=f"{isotope}-{self_shielding}")
                
                dose = fix_spikes(dose, material)
                
                plt.plot(dose[:,0], dose[:,1], ls, color=color)            
                
                results.setdefault(isotope, {})
                results[isotope].setdefault(self_shielding, {})
                results[isotope][self_shielding]["h(10) [uSv/h per MBq/m^2]"] = h_10
                results[isotope][self_shielding][material] = dose[:,:2]
        if 1:
            pass
            #plt.ylim([1e-9, 2])
            plt.yscale('log')
        else:
            if material == "Lead":
                plt.xlim([-0.1, 3])
            else:
                plt.xlim([-1, 40])
        #plt.xlim([-0.1, 3])
        #plt.ylim([1E-8, 2])
        plt.title(f"{material}")    
        plt.legend(loc="lower right")
        #plt.show()
    
    for dupl in DUPLICATE_RESULTS:
        results[dupl[1]] = results[dupl[0]]
    
    with open(PICKLE_FILE, 'wb') as f:
        pickle.dump(results, f)



from functools import partial

from PyQt5.QtWidgets import  (QApplication, QGraphicsView,  
                              QMainWindow, QGraphicsScene)

from PyQt5.QtGui import QPainter, QTransform
from PyQt5.QtCore import Qt, pyqtSignal, QRectF, QPointF
from pyrateshield.gui.legend import Legend
from pyrateshield.model import Model


from pyrateshield import labels
from pyrateshield.model_items import (Wall, SourceNM, SourceCT, 
                                      SourceXray, CriticalPoint, Ruler)
from pyrateshield.gui.graphics_items import (GraphicsModelItem,                                              
                                             GraphicsSourceNM, 
                                             GraphicsSourceCT, 
                                             GraphicsSourceXray, 
                                             DosemapPixmap,                                             
                                             GraphicsWall, 
                                             GraphicsContourItem,
                                             FloorplanPixmap,
                                             Marker,
                                             Cross,
                                             RulerLine,
                                             GraphicsItem)


        


class GraphicsScene(QGraphicsScene):
    _pixel_size_line = None
    _floorplan = None
    _dosemap = None
    

        
    def _select_all(self, select=True):
       self.blockSignals(True)
       for item in self.graphicsModelItems() + self.markers():
           item.setSelected(select)
       self.blockSignals(False)
       self.selectionChanged.emit()
    
    def removeItem(self, item):
        if item is None or item not in self.items():
            return
        super().removeItem(item)
       
    def addItem(self, item):        
        if isinstance(item, FloorplanPixmap):
            self.removeFloorplan()
        elif isinstance(item, DosemapPixmap):
            self.removeDosemap()
        super().addItem(item)
        
    def silentSelectItem(self, item, deselect_others=True):
        self.blockSignals(True)
        for selected in self.selectedItems():
            if item is not selected:
                selected.setSelected(False)
                
        if item not in self.selectedItems():
             item.setSelected(True)
        self.blockSignals(False)
        #self.selectionChanged.emit()
                
    def sendToForeground(self, item):
        if isinstance(item , Marker):
            item = item.parentItem()
        
        for gitem in self.graphicsModelItems():
            if gitem is item:
                gitem.setZValue(100)
            else:               
                gitem.setZValue(item.style_zvalue())
        
       
                
    
    
    def deselectAll(self):     
        self._select_all(False)
   
    def selectAll(self):
        self._select_all(True)
        
    def floorplan(self):
        items = [item for item in self.items()\
                 if isinstance(item, FloorplanPixmap)]
        if len(items) == 1:
            return items[0]
        if len(items) == 0:
            return None
    
    def dosemap(self):
        items = [item for item in self.items()\
                 if isinstance(item, DosemapPixmap)]
        return None if len(items) == 0 else items[0]
        
    def legend(self):
        items = [item for item in self.items()\
                 if isinstance(item, Legend)]
        return None if len(items) == 0 else items[0]
        
    
    def removeFloorplan(self):
        self.removeItem(self.floorplan())
        
    def removeDosemap(self):
        self.removeItem(self.dosemap())
                
    def removeContours(self):
        for item in self.contourItems():
            self.removeItem(item)
                
    def contourItems(self):
        return [item for item in self.items()\
                if isinstance(item, GraphicsContourItem)]    
            
    def graphicsModelItems(self):
        return [item for item in self.items()\
                if isinstance(item, GraphicsModelItem)]
            
    def walls(self):
        return [item for item in self.items()\
                if isinstance(item, GraphicsWall)]
            
    def markers(self):
        return [item for item in self.items()\
                if isinstance(item, Marker)]
            
    def rulers(self):
        return [item for item in self.items()\
                if isinstance(item, RulerLine)]
                
    def setVisibleGraphicsModelItems(self, visible):
        for item in self.graphicsModelItems():            
                item.setVisible(visible)
    
    def itemsInRect(self, rect):
        return [item for item in self.graphicsModelItems() + self.markers()\
                if self.itemInRect(rect, item)]
            
    def itemInRect(self, rect, item):
        pos = item.pos()
        
        return pos.x() >= rect.left() and pos.x() <= rect.right()\
                and pos.y() <= rect.bottom() and pos.y() >= rect.top()
                
    def selectedGraphicsModelItems(self):   
        selected=set()
        for item in self.selectedItems():
            if isinstance(item, Marker):
                item = item.parentItem()
            if isinstance(item, GraphicsModelItem):
                selected.add(item)
        return list(selected)
                
                
    def updateStyles(self):
        for item in self.graphicsModelItems() + self.contourItems():
            item.updateStyle()
            
            
    def sceneUnitsToPoints(self, scene_units):
        dpi = QApplication.desktop().physicalDpiX()
        
        scale_factor = self.width() / self.sceneRect().width()
            
        pixels = scene_units * scale_factor
        
        points = pixels * 72 / dpi
        return points
    
    def pointsToSceneUnits(self, points):
        dpi = QApplication.desktop().physicalDpiX()
        
        # Convert points to pixels
        pixels = points * dpi / 72.0
    
       
        scale_factor = self.width() / self.sceneRect().width()
        
        
    
        # Convert pixels to scene units
        scene_units = pixels / scale_factor
        return scene_units
    
    

class Graphics(QGraphicsView):
    mouse_press_point = None
    _dosemap = None
    _zoomRect = None
    _ctrl_pressed = False


    leftMouseClick      = pyqtSignal(object, object)
    rightMouseClick     = pyqtSignal(object, object)
    mouseMove           = pyqtSignal(object)
    mouseRelease        = pyqtSignal(object)
    askTooltip          = pyqtSignal(object)
    
    

    pixel_item_class_map = {SourceNM:       GraphicsSourceNM,
                            SourceXray:     GraphicsSourceXray,
                            SourceCT:       GraphicsSourceCT,
                            CriticalPoint:  Cross,
                            Wall:           GraphicsWall,
                            Ruler:          RulerLine}
    
        
    def __init__(self, model=None, scene=None, parent_controller=None):
        if model is None:
            model = Model()
        
        if scene is None:
            scene = GraphicsScene()
                    
        super().__init__(scene)
        
        

        self.setRenderHints(QPainter.HighQualityAntialiasing)
        self.setModel(model)
        self.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)
        self.scene().selectionChanged.connect(self.selection_changed)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        
        self.last_clicked_item = None
    
    def ctrlPressed(self):
        return self._ctrl_pressed
    
    def selection_changed(self):
       
        self.model.history.clean()
        
    def visibleRect(self):
        return self.mapToScene(self.viewport().rect()).boundingRect()
    
    def selectedModelItems(self):
        items = self.scene().selectedItems()
        
        points = [item.model for item in items\
                  if not isinstance(item, Marker)]
            
        lines = []
        for item in items:
            if isinstance(item, Marker):
                if item.parent_line not in lines:
                    lines += [item.parent_line.model]
                        
        return self.model.sort_model_items(points + lines)
    
    def wheelEvent(self, event):
        if event.modifiers() != Qt.ControlModifier: return
        zoom_in_factor = 1.05
        zoom_out_factor = 0.95
        zoom_steps = 10

        angle = event.angleDelta().y() / 8
        steps = angle / 15

        if steps > 0:
            for _ in range(zoom_steps):
                zoom_factor = self.zoomFactor() * zoom_in_factor
        else:
            for _ in range(zoom_steps):
                zoom_factor = self.zoomFactor() * zoom_out_factor
        
        zoom_factor = max(0.1, min(zoom_factor, 50.0))
        
        self.setTransform(QTransform().scale(zoom_factor, zoom_factor))

        if steps < 0 and  self.visibleRect().contains(self.sceneRect()):
            self.fitInView(self.sceneRect(), Qt.KeepAspectRatio)
        
        self.zoomChanged()
    
    def mouseMoveEvent(self, event):
        self.mouseMove.emit(event)
        super().mouseMoveEvent(event)
        
    def mouseReleaseEvent(self, event):    
        self.mouseRelease.emit(event)

        super().mouseReleaseEvent(event)
        self.model.history.clean()
        self.mouse_press_point = None
    
    def mousePressPoint(self):
        if self.mouse_press_point is not None:
            return self.mouse_press_point
        else:
            return QPointF()
    
    def selectionToModel(self):        
        for item in self.scene().selectedGraphicsModelItems():  
               item.graphicsPositionToModel()
        
    def mousePressEvent(self, event):
        self.mouse_press_point = self.mapToScene(event.pos())
        self.last_clicked_item
        
        item = self.itemAt(event.pos())
        
        for scene_item in self.scene().graphicsModelItems():
            if scene_item is item or item is None:
                continue
            
            if item.zValue() == 100:

                item.updateStyle()
            
            if item.zValue() > 50:
                raise 
                
        if event.button() == 1:  # Left mouse button
            self.leftMouseClick.emit(event, item)
                           
        elif event.button() == 2:  # Right mouse button
            self.rightMouseClick.emit(event, item)
        
        super().mousePressEvent(event)
        
        
    def setModel(self, model):                  
        self.model = model
        

        for sequence in (self.model.sources_nm,
                         self.model.sources_ct,
                         self.model.sources_xray,
                         self.model.critical_points,
                         self.model.walls,
                         self.model.rulers):
            
            cbk = partial(self.model_item_about_to_be_removed, sequence)
            sequence.rowsAboutToBeRemoved.connect(cbk)
            
            cbk = partial(self.model_item_inserted, sequence)
            sequence.rowsInserted.connect(cbk)
        
        cbk = self.refresh
        self.model.floorplan.update_event.connect(self.refresh)
        
        self.refresh()
    
   
    
    def refresh(self, _=None):
        self.model.dosemap.result = None        
        self.scene().clear()
        
        floorplan = FloorplanPixmap(model=self.model.floorplan)
        self.scene().addItem(floorplan)
        
        dosemap = DosemapPixmap(style=self.model.dosemap_style)                               
        self.scene().addItem(dosemap)
        
        self.create_pixel_items()
        self.scene().updateStyles()
        self.resetZoom()
        
        
    def model_item_inserted(self, sequence, index):
        model_item = sequence.itemAtIndex(index)
       
        self.create_pixel_item(model_item, select=True)

    def model_item_about_to_be_removed(self, sequence, index):

        model_item = sequence.itemAtIndex(index)
        
        pixel_item = self.pixel_item_for_model_item(model_item)
        
        if pixel_item is self.last_clicked_item:
            self.last_clicked_item = None
       
        if pixel_item is not None:            
            self.scene().removeItem(pixel_item)        
            del pixel_item
            
        
    def floorplan_update(self, event_data):
        if event_data[1] == labels.GEOMETRY:
            pixel_size_cm = self.model.floorplan.geometry.pixel_size_cm
            for item in self.scene().graphicsModelItems():
                if item is not self.pixelSizeLine():
                    item.setPixelSizeCM(pixel_size_cm)
                    
        # elif event_data[1] == labels.IMAGE:
        #     self.model.dosemap.result = None
        #     self.load_floorplan()

    
    def pixel_item_for_model_item(self, model_item):
        if isinstance(model_item, GraphicsModelItem):
            return model_item
        
        for pixel_item in self.scene().graphicsModelItems():                                       
            if pixel_item.model is model_item:
                return pixel_item


    def create_pixel_items(self):
        for model_item in [*self.model.sources_nm,
                           *self.model.sources_ct,
                           *self.model.sources_xray,
                           *self.model.walls,
                           *self.model.critical_points,
                           *self.model.rulers]:
            self.create_pixel_item(model_item, select=False)
            
        
    def create_pixel_item(self, model_item, select=True):
      
        pixel_item_class  = self.pixel_item_class_map[model_item.__class__]
        pixel_size_cm = self.model.floorplan.geometry.pixel_size_cm
        
        pixel_item = pixel_item_class(model=model_item, pixel_size_cm=pixel_size_cm)
      
        self.scene().addItem(pixel_item)
        
        return pixel_item

    

    def zoomFactor(self):
        return self.transform().m11()
   
    def zoomChanged(self):
        for wall in self.scene().walls():
            wall.updateLineStyle()
        for contour in self.scene().contourItems():
            contour.updateStyle()
        for line in self.scene().rulers():
            line.updateStyle()
 
    
    def setZoomRect(self, rect):       
        self.fitInView(rect, Qt.KeepAspectRatio)
        self.zoomChanged()
        
    def resetZoom(self):
        self.setZoomRect(QRectF(self.scene().floorplan().pixmap().rect()))

    def resizeEvent(self, *args):
        super().resizeEvent(*args)
        if self.scene().floorplan() is not None:
            rect = QRectF(self.scene().floorplan().pixmap().rect())
            if self.visibleRect().contains(rect):
                self.resetZoom()
  
    def legend(self):
        if self.scene().contourItems() and self.model.dosemap_style.show_legend:
            return self.scene().dosemap().legend()
        else:
            return None
  
    def paintEvent(self, event):
        # Call the base class paintEvent to ensure proper drawing
        super().paintEvent(event)
        if self.legend() is not None:
            self.paintLegend()
            
    def paintLegend(self):
        p = QPainter(self.viewport())
        p.setRenderHints(self.renderHints())
        legend = self.legend()
        rect = QRectF(legend.left(), legend.top(),
                      legend.sceneRect().width(),
                      legend.sceneRect().height())
        
        legend.render(p, rect)
        
    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Z and event.modifiers() == Qt.ControlModifier:
            self._ctrl_pressed = False
            self.model.history.undo()
        elif event.key() == Qt.Key_Y and event.modifiers() == Qt.ControlModifier:
            self._ctrl_pressed = False
            self.model.history.redo()
        elif event.key() == Qt.Key_Left:
            self.moveSelectedItems(-1, 0)
        elif event.key() == Qt.Key_Right:
            self.moveSelectedItems(1, 0)
        elif event.key() == Qt.Key_Up:
            self.moveSelectedItems(0, -1)
        elif event.key() == Qt.Key_Down:
            self.moveSelectedItems(0, 1)
        elif event.key() == Qt.Key_Control:
            self._ctrl_pressed = True
            
        super().keyPressEvent(event)
        
    def keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Control:
            self._ctrl_pressed = False
        super().keyReleaseEvent(event)

    def moveSelectedItems(self, dx, dy):
        MOVE_PIXELS = 10
        dx *= MOVE_PIXELS
        dy *= MOVE_PIXELS
        
        for item in self.scene().selectedGraphicsModelItems():
            item.model.move_pixels(dx, dy)
        
        
        
        

if __name__ == '__main__':
    
    app = QApplication([])
    
    
    project = '../../example_projects/SmallProject/project.zip'
    project = '../../example_projects/SmallProject/project.zip'
    model = Model.load_from_project_file(project)
    window = QMainWindow()
    
    view = Graphics(model=model)
    window.resize(1920, 1080)
    
   
    window.setCentralWidget(view)
    
    # view.legend.setBottomRight()
   
    
   
    

    
    window.show()
    
    
  

    app.exec_()



import os
import pandas as pd
import numpy as np
from pyrateshield import labels

#ENERGY_keV = 'Energy [keV]'
MFP = 'mfp'
MASS_ATTENUATION = 'mu/p [cm^2/g]'
ENERGY_MeV = 'Energy [MeV]'

folder = os.path.dirname(__file__)

file = os.path.join(folder, 'buildup.xls')
BUILDUP_TABLES = pd.read_excel(os.path.join(folder, file), sheet_name=None)

file = os.path.join(folder, 'attenuation.xls')
ATTENUATION_TABLES = pd.read_excel(os.path.join(folder, file), sheet_name=None)

ATTENUATION = {}


# for material, table in ATTENUATION_TABLES.items():
#     if material == 'References':
#         continue
    
#     ATTENUATION[material] = {}
#     ATTENUATION[material][ENERGY_MeV] = np.array(table[ENERGY_MeV])
#     ATTENUATION[material][MASS_ATTENUATION] = np.array(table[MASS_ATTENUATION])


# BUILDUP = {}

# for material, table in ATTENUATION_TABLES.items():
#     if material == 'References':
#         continue
    
#     ATTENUATION[material] = {}
#     ATTENUATION[material][ENERGY_MeV] = np.array(table[ENERGY_MeV])
#     ATTENUATION[material][MASS_ATTENUATION] = np.array(table[MASS_ATTENUATION])
    
    

        




  


import multiprocessing
from multiprocessing.sharedctypes import Value as mpValue
import ctypes
import queue
import time
import psutil
from pyrateshield.constants import CONSTANTS
from pyrateshield import labels
from pyrateshield.logger import Logger, GLOBAL_LOG_LEVEL
from PyQt5.QtCore import QObject, pyqtSignal
from pyrateshield.radtracer import radtracer
from pyrateshield.pyshield.engine import Engine
import pandas as pd
from scipy.ndimage import zoom

NO_WORK = 0
DOSEMAP_WORK = 1
CRITICAL_POINT_WORK = 2


MAX_CORES = 8
LOG_LEVEL = Logger.LEVEL_DEBUG

UNSUPPORTED_PYSHIELD_ISOTOPES = ['Y-90']
CRITICAL_POINT_NAME         = 'Critical Point Name'
RADTRACER_DOSE              = 'RadTracer Dose [mSv]'
PYSHIELD_DOSE               = 'PyShield Dose [mSv]'
OCCUPANCY_FACTOR            = 'Occupancy factor'
RADTRACER_DOSE_CORRECTED    = 'RadTracer Dose corrected for occupancy [mSv]'
PYSHIELD_DOSE_CORRECTED     = 'PyShield Dose corrected for occupancy [mSv]'
SOURCE_NAME                 = 'Source Name'

COLUMNS = [CRITICAL_POINT_NAME, PYSHIELD_DOSE, RADTRACER_DOSE, 
           OCCUPANCY_FACTOR, PYSHIELD_DOSE_CORRECTED, RADTRACER_DOSE_CORRECTED]

SPLIT_COLUMNS = [CRITICAL_POINT_NAME, SOURCE_NAME, PYSHIELD_DOSE, 
                 RADTRACER_DOSE, OCCUPANCY_FACTOR, 
                 RADTRACER_DOSE_CORRECTED, PYSHIELD_DOSE_CORRECTED]

def resize_array(source_array, target_shape):
    # Need faster method for painting dosemap on top of the floorplan.
    # Obtain same size by interpolation
    zoom_factors = [float(t) / s for s, t in zip(source_array.shape, target_shape)]
    resized_array = zoom(source_array, zoom_factors, order=1)
    return resized_array

def is_supported_by_pyshield(src):
    supported = CONSTANTS.pyshield_supported_isotopes
    return src.label == labels.SOURCES_NM and src.isotope in supported


def is_supported_by_radtracer(src):
    supported = CONSTANTS.radtracer_supported_isotopes
    return src.label != labels.SOURCES_NM or src.isotope in supported


def get_radtracer_dosemap(source, project, pyshield_engine):
    if is_supported_by_radtracer(source):
        return radtracer.dosemap_single_source(source, project)
    elif is_supported_by_pyshield(source): # fallback to pyshield
        print(f'Source {source.name} not supported by radtracer, will use pyshield')   
        return pyshield_engine.source_dosemap(source)

def get_pyshield_dosemap(source, project, pyshield_engine):
    if is_supported_by_pyshield(source):    
        return pyshield_engine.source_dosemap(source)
    elif is_supported_by_radtracer(source): # fallback to radtracer
        # Xray will go through radtracer any how no need to warn
        if source.label == labels.SOURCES_NM:
            print(f'Source {source.name} not supported by pyshield, will use radtracer')
        return radtracer.dosemap_single_source(source, project)
    
def get_dosemap(source, project, pyshield_engine=None):
    if pyshield_engine is None:
        pyshield_engine = Engine.from_pyrateshield(project)
    # Pyshield set as engine
    if project.dosemap.engine == labels.PYSHIELD:
        return get_pyshield_dosemap(source, project, pyshield_engine=pyshield_engine)
    # Radtracer set as engine
    elif project.dosemap.engine == labels.RADTRACER:
        return get_radtracer_dosemap(source, project, pyshield_engine=pyshield_engine)

def empty_critical_point_format():
    return pd.DataFrame([{CRITICAL_POINT_NAME:      None,
                          SOURCE_NAME:              None,
                          RADTRACER_DOSE:           None,
                          PYSHIELD_DOSE:            None,
                          OCCUPANCY_FACTOR:         None,
                          RADTRACER_DOSE_CORRECTED: None,
                          PYSHIELD_DOSE_CORRECTED:  None
                          }])

def source_critical_point_format(source, 
                                 critical_point, 
                                 dose_radtracer, 
                                 dose_pyshield):
    
    factor = critical_point.occupancy_factor
    
    return {CRITICAL_POINT_NAME:        critical_point.name,
            SOURCE_NAME:                source.name,
            RADTRACER_DOSE:             dose_radtracer,
            PYSHIELD_DOSE:              dose_pyshield,
            OCCUPANCY_FACTOR:           factor,
            RADTRACER_DOSE_CORRECTED:   factor * dose_radtracer,
            PYSHIELD_DOSE_CORRECTED:    factor * dose_pyshield}
                          
def get_critical_point_for_source(project, critical_point, source,
                                  engine=labels.RADTRACER,
                                  pyshield_engine=None):
    
    if engine == labels.PYSHIELD and pyshield_engine is None:
        pyshield_engine = Engine.from_pyrateshield(project)
    
    def get_radtracer_dose(source):
        return radtracer.pointdose_single_source(
            critical_point.position, source, project)
    
    def get_pyshield_dose(source):
        return pyshield_engine.dose_at_point(
                critical_point.position, sources=[source])
    
    if not is_supported_by_pyshield(source)\
        and not is_supported_by_radtracer(source):
        raise RuntimeError('Unsupported source!')
        
    if engine == labels.PYSHIELD and is_supported_by_pyshield(source):
        dose = get_pyshield_dose(source)
    elif engine == labels.PYSHIELD:
        dose = get_radtracer_dose(source)
    elif engine == labels.RADTRACER and is_supported_by_radtracer(source):
        dose = get_radtracer_dose(source)
    elif engine == labels.RADTRACER:
        dose = get_pyshield_dose(source)
    return dose
    

def get_critical_point(project, critical_point, 
                       engine=labels.RADTRACER,
                       pyshield_engine=None,
                       sum_dose = True):
    
    sources = Dosemapper.project_sources(project)
    
    total_dose = {}
    for source in sources:
        dose = get_critical_point_for_source(project, critical_point, source, 
                                             engine=engine, 
                                             pyshield_engine=pyshield_engine)
        
        total_dose[(source.label, source.name)] = dose
    
    if sum_dose:
        total_dose = sum(total_dose.values())
        
    return total_dose


def get_formatted_critical_point_result(project, critical_point, source, 
                                        pyshield_engine=None):
    
    dose_pyshield = get_critical_point_for_source(
        project, critical_point, source,
        engine=labels.PYSHIELD,
        pyshield_engine=pyshield_engine)
        
    dose_radtracer = get_critical_point_for_source(
        project, critical_point, source,
        engine=labels.RADTRACER,
        pyshield_engine=pyshield_engine)
   
    formatted = source_critical_point_format(
        source, critical_point, 
        dose_radtracer=dose_radtracer,
        dose_pyshield=dose_pyshield)
    
    return formatted
    
class Worker(multiprocessing.Process):
    def __init__(self, project_queue, work_queue, 
                 source_queue, workt_type, update_flag, stop_flag):
        multiprocessing.Process.__init__(self)
        
        
        self.project_queue = project_queue
        
        self.source_queue = source_queue
        
        self.work_type = workt_type
        
        self.work_queue = work_queue
       
        self.update_flag = update_flag
        
        self.stop_flag = stop_flag      
        
    
    def update_project(self):
        
        if self.update_flag.value:        
        
            self.project = self.project_queue.get(timeout=2)
      
            self.pyshield_engine = Engine.from_pyrateshield(self.project)
        
        self.update_flag.value = False
        
    def run(self):
        while not self.stop_flag.value:  
            work_type = self.work_type.value
            if self.update_flag.value:
                self.update_project()
                
            if work_type == NO_WORK:
                time.sleep(0.1)
               
            elif work_type > NO_WORK and not self.source_queue.empty(): 
                try:
                    sources = self.source_queue.get(timeout=0.1)
                    # work type changed during time out 
                    if work_type != self.work_type.value:
                        self.source_queue.put(sources)
                        continue
                except queue.Empty:
                    continue

                
                for source in sources:
                    if work_type == DOSEMAP_WORK:
                        self.work_queue.put(get_dosemap(source, self.project))
                    elif work_type == CRITICAL_POINT_WORK:
                        self.work_queue.put(self.get_critical_points(source))
                    else:
                        raise RuntimeError
                
        
            
    def get_critical_points(self, source):
        results = []
        for critical_point in self.project.critical_points:
            result = get_formatted_critical_point_result(self.project, critical_point, source,
                                                pyshield_engine=self.pyshield_engine)
            results.append(result)
        return results
    

   

class Dosemapper(QObject):
    _workers = None
    _work_queue = None
    _max_cpus = MAX_CORES
    _source_queue = None
    _cpus = None
    progress = pyqtSignal(float)
    calculation_time = pyqtSignal(float)
    def __init__(self, multi_cpu=True):
        super().__init__()
        if not multi_cpu:
            self.set_cpus(1)
    
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._stop_workers(self._workers)
       
    def source_queue(self):
        if self._source_queue is None:
            self._source_queue = multiprocessing.Queue()
        return self._source_queue
    
    def work_queue(self):
        if self._work_queue is None:
            self._work_queue  = multiprocessing.Queue()
        return self._work_queue
    
    def set_cpus(self, cpus=None):
        if self._cpus != cpus:
            self._cpus = cpus
            
    def cpus(self):
        if self._cpus is None:
            max_cpus = psutil.cpu_count(logical=False)
            self._cpus = min(max_cpus, self._max_cpus)
            self._cpus = max(self._cpus, 1)
        return self._cpus
    
    def update_workers(self, project=None, work_type=None):
        
        while len(self.workers()) < self.cpus():
            self.workers().append(self._new_worker())
            
        while len(self.workers()) > self.cpus():
           
            self._stop_worker(self.workers()[0])
            
            self._workers.remove(self._workers[0])
        
        while any(w.update_flag.value for w in self.workers()):
            time.sleep(0.1)
        
        if project is not None:
            for w in self.workers():
                w.project_queue.put(project)
                w.work_type.value = work_type if work_type is not None else 0
                w.update_flag.value = True
        
        if work_type is not None:
            for w in self.workers():
                w.work_type.value = work_type
            
        return self.workers()
    
    def workers(self):
        if self._workers is None:
            self._workers = []
        return self._workers
    
    def _new_worker(self):
        project_queue   = multiprocessing.Queue()
        stop_flag       = mpValue(ctypes.c_bool, False)
        update_flag     = mpValue(ctypes.c_bool, False)
        work_type       = mpValue(ctypes.c_int8, NO_WORK)
        args = (
            project_queue,
            self.work_queue(),
            self.source_queue(),
            work_type,
            update_flag,
            stop_flag,
        )
        p = Worker(*args)            
        p.start()
        return p

    def _stop_worker(self, worker):
        worker.stop_flag.value = True
        worker.join()
        time.sleep(0.1)
        if worker.is_alive():
            worker.close()
    
    def _stop_workers(self, workers):
        while workers is not None and len(workers) > 0:
            self._stop_worker(workers[0])
            workers.remove(workers[0])

   
    @staticmethod
    def project_sources(project):
        sources = list(project.sources_nm) + list(project.sources_ct)\
            + list(project.sources_xray)
        return [source for source in sources if source.enabled]
        
        
    @staticmethod
    def get_single_cpu_dosemap(project, sources=None):
        # DEBUGGING Purpuses
    
                
        dosemap = None        
        update_dosemap = lambda dm: dm if dosemap is None else (dosemap + dm)
            
        
        sources = Dosemapper.project_sources(project)
        
        pyshield_engine = Engine.from_pyrateshield(project)
         
        for source in sources:
            source_dosemap = get_dosemap(source, project, pyshield_engine=pyshield_engine)
        
            dosemap = update_dosemap(source_dosemap)
        
        if project.dosemap_style.interpolate:
            dosemap = resize_array(dosemap, project.floorplan.image.shape)
            
        return dosemap
 
    def get_critical_point(self, project, sources=None):
        if sources is None:
            sources = self.project_sources(project)
        
        src_pos_dct = {}

        for source in sources:
            # Make a dictionary of source_positions and group the corresponding sources
            # I.e.: { (x,y,z): [src1, src2, src3 ], (x,y,z): [src4, src5, ... ] }
            src_pos_dct.setdefault(tuple(source.position), []).append(source)
            
        sources_grouped = [srcs for srcs in src_pos_dct.values()]
        
        self.update_project(project)
        
        for grp in sources_grouped:
            self.source_queue().put(grp)
            
        
    def get_dosemap(self, project, sources=None):
        #print(f'Using engine: {project.dosemap.engine}')
        
        
        if self.cpus() == 1:
            return self.get_single_cpu_dosemap(project, sources)
        
        start = time.time()
        self.update_workers(project=project)
        
        
        
        if sources is None:
            sources = self.project_sources(project)
     
        
        # Check if there is any work to be done
        if not len(sources):
            return None

        
        # Sources with the same position are processed by the same worker.
        # This allows PyShield to benefit from caching.
        # MS: Group sources by default won't affect radtracer
            
        src_pos_dct = {}

        for source in sources:
            # Make a dictionary of source_positions and group the corresponding sources
            # I.e.: { (x,y,z): [src1, src2, src3 ], (x,y,z): [src4, src5, ... ] }
            src_pos_dct.setdefault(tuple(source.position), []).append(source)
            
        sources_grouped = [srcs for srcs in src_pos_dct.values()]

        
        self.update_workers(work_type=DOSEMAP_WORK)
        
        for grp in sources_grouped:
            self.source_queue().put(grp)
            
        
        # Collect the results. Make sure to get as many results from the queue
        # as the number of sources
        nr_of_sources = sum(len(grp) for grp in sources_grouped)
                
        dosemap = None        
        update_dosemap = lambda dm: dm if dosemap is None else (dosemap + dm)
       

        self.progress.emit(0)
        for i in range(nr_of_sources):
            dosemap = update_dosemap(self.work_queue().get())
            self.progress.emit(i/nr_of_sources)
        
        stop = time.time()
        
        self.progress.emit(1)
        self.calculation_time.emit(stop-start)
        self.update_workers(work_type=NO_WORK)
        
        if not self.work_queue().empty(): raise RuntimeError()
        
        if project.dosemap_style.interpolate:
            dosemap = resize_array(dosemap, project.floorplan.image.shape)
       
        return dosemap        
    
    def get_critical_points(self, project, sum_sources=True):
    
        
        
        self.update_workers(project=project)
       
        
        sources = self.project_sources(project)
        
        
        if not len(sources):
            return empty_critical_point_format()
       
        
        if self.cpus() == 1:
            results = []
            pyshield_engine = Engine.from_pyrateshield(project)
            for source in sources: 
                for critical_point in project.critical_points:
                    result = get_formatted_critical_point_result(project, critical_point, source,
                                                                 pyshield_engine=pyshield_engine)
                    results += [result]
        else:
            start = time.time()
            
            src_pos_dct = {}
            
            for source in sources:
                # Make a dictionary of source_positions and group the corresponding sources
                # I.e.: { (x,y,z): [src1, src2, src3 ], (x,y,z): [src4, src5, ... ] }
                src_pos_dct.setdefault(tuple(source.position), []).append(source)
           
            sources_grouped = [srcs for srcs in src_pos_dct.values()]
            
            
            
            for grp in sources_grouped:
                self.source_queue().put(grp)
            
            
            
            nr_of_sources = sum(len(grp) for grp in sources_grouped)
            
            
           
            
            
            results = None
            
            update_results = lambda result: result if results is None else (results+result)
          
            self.progress.emit(0)
            
            self.update_workers(work_type=CRITICAL_POINT_WORK)
           
            for i in range(nr_of_sources):
                result = self.work_queue().get()
                
                results = update_results( result)    
                
                self.progress.emit(i/nr_of_sources)
            stop = time.time()
            self.calculation_time.emit(stop-start)
            self.progress.emit(1)
            self.update_workers(work_type=NO_WORK)
            
        report = pd.DataFrame(results)[SPLIT_COLUMNS]
        
        if sum_sources:
            aggfunc = {RADTRACER_DOSE: sum, RADTRACER_DOSE_CORRECTED: sum,
                       PYSHIELD_DOSE: sum, PYSHIELD_DOSE_CORRECTED: sum,
                       OCCUPANCY_FACTOR: max}
                
            report = pd.pivot_table(report, index=CRITICAL_POINT_NAME, 
                                            aggfunc=aggfunc)
            
            # preserve order
            pnames = [crit_point.name for crit_point in project.critical_points\
                      if crit_point.enabled]
                
            report = report.reindex(pnames)
            
            report = report.reset_index()
           
            report = report[COLUMNS]
            
       
        
        return report
            
    
    # @staticmethod
    # def get_sources_critical_points(project, sum_sources=True):
    #     sources = list(project.sources_nm) + list(project.sources_ct) + list(project.sources_xray)
        
    #     # Exclude sources which are not set to Enabled
    #     sources = [src for src in sources if src.enabled]
        
    #     # Exclude points which are not set to Enabled
    #     crit_points = [crp for crp in project.critical_points if crp.enabled]
                
    #     if len(crit_points) == 0 or len(sources) == 0:            
    #         return Dosemapper.empty_critical_point_result()
        
    #     reports = []        
        
    #     pyshield_engine = Engine.from_pyrateshield(project)
        
    #     for crit_point in crit_points:            
    #         for source in sources:
    #             if is_supported_by_radtracer(source):
    #                 dose_radtracer = radtracer.pointdose_single_source(
    #                     crit_point.position, source, project)
    #             else:
    #                 # At least one isotope not supported by radtracer
    #                 dose_radtracer = None
                
    #             if is_supported_by_pyshield(source):
    #                 dose_pyshield = pyshield_engine.dose_at_point(crit_point.position,
    #                                                      sources=[source])
    #             else:
    #                 dose_pyshield = None
                    
    #             if dose_radtracer is None and dose_pyshield is None:
    #                 raise RuntimeError
    #             elif dose_radtracer is None:
    #                 dose_radtracer = dose_pyshield
    #             elif dose_pyshield is None:
    #                 dose_pyshield = dose_radtracer
                    
                
    #             reports += [Dosemapper.get_source_critical_point_result(
    #                 source, crit_point, dose_radtracer, dose_pyshield)]
    #     report = pd.DataFrame(reports)[SPLIT_COLUMNS]
        
    #     if sum_sources:
    #         aggfunc = {RADTRACER_DOSE: sum, RADTRACER_DOSE_CORRECTED: sum,
    #                    PYSHIELD_DOSE: sum, PYSHIELD_DOSE_CORRECTED: sum,
    #                    OCCUPANCY_FACTOR: max}
                
    #             report = pd.pivot_table(report, index=CRITICAL_POINT_NAME, 
    #                                             aggfunc=aggfunc)
                
    #             # preserve order
    #             pnames = [crit_point.name for crit_point in project.critical_points\
    #                       if crit_point.enabled]
                    
    #             report = summed_report.reindex(pnames)
                
    #             report = summed_report.reset_index()
               
    #             report = summed_report[COLUMNS]
                
        
    #     return report
        
    # @staticmethod
    # def get_critical_points(project):
        
        
    #     if len(project.sources_nm) + len(project.sources_ct) + len(project.sources_xray) == 0\
    #         or len(project.critical_points) == 0:
    #             return Dosemapper.empty_critical_point_result()
            
    #     report = Dosemapper.get_sources_critical_points(project)
        
    #     aggfunc = {RADTRACER_DOSE: sum, RADTRACER_DOSE_CORRECTED: sum,
    #                PYSHIELD_DOSE: sum, PYSHIELD_DOSE_CORRECTED: sum,
    #                OCCUPANCY_FACTOR: max}
        
    #     summed_report = pd.pivot_table(report, index=CRITICAL_POINT_NAME, 
    #                                    aggfunc=aggfunc)
        
    #     # preserve order
    #     pnames = [crit_point.name for crit_point in project.critical_points\
    #               if crit_point.enabled]
            
    #     summed_report = summed_report.reindex(pnames)
        
    #     summed_report = summed_report.reset_index()
       
    #     summed_report = summed_report[COLUMNS]
    #     return summed_report

    
        
            
if __name__ == "__main__":
    from pyrateshield.model import Model
    import timeit
    import time
    #model = Model.load_from_project_file('/Users/marcel/git/pyrateshield/example_projects/LargeProject/large_project.psp')
    #model = Model.load_from_project_file('../example_projects/LargeProject/project.zip')
    model = Model.load_from_project_file('../example_projects/SmallProject/project.zip')
    #model = Model.load_from_project_file('../example_projects/Stabin/Tb-161.zip')
    #model = Model.load_from_project_file('../example_projects/Lu-177.psp')
    model.dosemap.grid_matrix_size = 120
    
    with Dosemapper(multi_cpu=True) as dm:   
        dm.progress.connect(print)
        model.dosemap.engine = labels.PYSHIELD
        #dm = dm.get_dosemap(model)
        #print("PyShield", timeit.timeit(lambda: dm.get_dosemap(model), number=1) )
        print("PyShield", timeit.timeit(lambda: dm.get_dosemap(model), number=1) )
    
        #time.sleep(1) 
        #print()
        
        model.dosemap.engine = labels.RADTRACER
        #print("Radtracer", timeit.timeit(lambda: dm.get_dosemap(model), number=1) )
        #print("Radtracer", timeit.timeit(lambda: dm.get_dosemap(model), number=1) )
        #results = dm.get_critical_points(model)
        print(timeit.timeit(lambda: dm.get_critical_points(model), number=1))

    # from pyshield import Sources
    # sources = Sources.from_pyrateshield(model)
    # print("Pyshield single CPU:", end=" ", flush=True)
    # print(timeit.timeit(lambda: sources.get_dosemap(), number=1) )
    
    
    
    





from pyrateshield import labels, __version__
from pyrateshield.gui.io import confirm_changes
from pyrateshield.gui.item_views import (EditPixelSizeView, 
                                         EditShieldingsView, 
                                         EditCriticalPointsView,  
                                         EditSourcesNMView,
                                         EditWallsView,
                                         EditSourcesCTView, 
                                         EditSourcesXrayView,
                                         EditMaterialsView,
                                         EditClearancesView)

from pyrateshield.gui.graphics import Graphics
from pyrateshield.gui.toolbar import Toolbar
                                         
                                      
                                      

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import (QWidget, QLabel, QToolBox, QTabWidget, 
                             QSplitter, QApplication, QPushButton, QTableView,
                             QVBoxLayout, QGridLayout, QCheckBox,  QMainWindow,
                             QProgressBar, QHBoxLayout, QStatusBar,
                             QSpacerItem, QSizePolicy)

#from pyrateshield.gui.mpl_view import NavigationToolbar, MplCanvas



# TOOLBOX_LAYOUT = {'Nuclear Medicine Sources':       [labels.SOURCES_NM, 
#                                                      labels.CLEARANCE],
#                   'Radiology Sources':              [labels.SOURCES_CT,
#                                                      labels.SOURCES_XRAY],
                                    
#                   'Walls, Shieldings && Materials':   [labels.WALLS,
#                                                       labels.SHIELDINGS,
#                                                       labels.MATERIALS],
#                   labels.CRITICAL_POINTS:           [labels.CRITICAL_POINTS],
#                   labels.PIXEL_SIZE_CM:             [labels.PIXEL_SIZE_CM]}                                                
                                              
                
TOOLBOX_LAYOUT = {'Nuclear Medicine Sources':       [labels.SOURCES_NM, 
                                                     labels.CLEARANCE],
                  'Radiology Sources':              [labels.SOURCES_CT,
                                                     labels.SOURCES_XRAY],
                                    
                  'Walls, Shieldings && Materials':   [labels.WALLS,
                                                      labels.SHIELDINGS,
                                                      labels.MATERIALS],
                  labels.CRITICAL_POINTS:           [labels.CRITICAL_POINTS],
                  
                  labels.PIXEL_SIZE_CM:             [labels.PIXEL_SIZE_CM]}
                    
            
   
RESULT_LAYOUT = [labels.CANVAS, labels.CRITICAL_POINT_REPORT_VIEW]

        
        
        
class ResultCriticalPointView(QWidget):
    check_button_label = 'Show contribution of each individual source'
    def __init__(self, parent=None):
        super().__init__(parent)
        self.create_widgets()
        self.create_layout()
        
    def create_widgets(self):
        self.critical_point_button        = QPushButton("Calculate Critical Points")
        self.save_critcial_point_button   = QPushButton("Save To Excel")
        
        self.table_view             = QTableView()
        self.table_view.verticalHeader().setVisible(False)
        self.source_checkbox = QCheckBox(self.check_button_label)

        #self.table_view.setRowCount(100)
        #self.table_view.setColumnCount(5)
        
    def create_layout(self):
        vlayout = QVBoxLayout()

        vlayout.addWidget(self.table_view)
    
        
        blayout = QGridLayout()
        blayout.addWidget(self.critical_point_button, 0, 0)
        blayout.addWidget(self.save_critcial_point_button, 0, 1)
        
        vlayout.addLayout(blayout)
        vlayout.addWidget(self.source_checkbox)
        
        self.setLayout(vlayout)
        
        # auto resize columns
        header = self.table_view.horizontalHeader()
        header.setSectionResizeMode(header.ResizeToContents)
        
        
    
        
class MainView(QMainWindow):
    _views = None
    focusSet = pyqtSignal(str)
    _currentFocus = None
    
    def __init__(self):
        super(MainView, self).__init__()
        self.views = self.create_widgets()
        self.create_layout()
        
        self.show()
        self.set_focus(labels.SOURCES_NM)

        self.setWindowTitle('PyrateShield (version ' + __version__ + ")")
    
    def closeEvent(self, event):
        if confirm_changes('Quit Application?'):
            super().closeEvent(event)
        else:
            event.ignore()
    
    
    def get_active_tool_panel_name(self):
        tool_index = self.toolbox.currentIndex()
        tool_name = self.toolbox.itemText(tool_index)
        if len(TOOLBOX_LAYOUT[tool_name]) > 1:
            tab_index = self.toolbox_tabs[tool_name].currentIndex()
        else:
            tab_index = 0
            
        tab_name = TOOLBOX_LAYOUT[tool_name][tab_index]
        
        return tab_name
    
    
    
    
    def set_focus(self, item_label):
        if self._currentFocus == item_label:
            return
        if item_label == labels.GEOMETRY:
            item_label = labels.PIXEL_SIZE_CM
        
        def location_in_toolbox(label):
            for name, items in TOOLBOX_LAYOUT.items():
                if label in items:
                    return name
            return None
            
        
        if item_label in RESULT_LAYOUT:
            self.result_container.setCurrentWidget(self.views[item_label])
        
        toolgroup = location_in_toolbox(item_label)
        
        if toolgroup is None:
            pass
            #raise ValueError(f'No view item with label {item_label}')
            
        else:
            toolbox_tab = self.toolbox_tabs[toolgroup]
            self.toolbox.setCurrentWidget(toolbox_tab)
            if len(TOOLBOX_LAYOUT[toolgroup]) > 1:
                toolbox_tab.setCurrentWidget(self.views[item_label])
                
       
        
        
            
    def set_tab_enabled(self, label, enabled):
        for toolbox_tab_name, tabs in TOOLBOX_LAYOUT.items():
            if label in tabs:
                index = tabs.index(label)
                tab_widget = self.toolbox_tabs[toolbox_tab_name]
                tab_widget.setTabEnabled(index, enabled)
                
            
    @property
    def status_text(self):
        return self.status_label.text()
    
    def set_status_text(self, text):
        self.status_text = text
    
    @status_text.setter
    def status_text(self, text):
        self.status_label.setText(str(text))
        
        
    
    def create_widgets(self):
        if self._views is None:
            views = {labels.SOURCES_NM:             EditSourcesNMView(parent=self),
                     labels.CLEARANCE:              EditClearancesView(parent=self),
                     labels.SOURCES_CT:             EditSourcesCTView(parent=self),
                     labels.SOURCES_XRAY:           EditSourcesXrayView(parent=self),
                    
                     labels.PIXEL_SIZE_CM:          EditPixelSizeView(parent=self),
                   
                     labels.CRITICAL_POINTS:        EditCriticalPointsView(),
                     labels.WALLS:                  EditWallsView(parent=self),
                     labels.SHIELDINGS:             EditShieldingsView(parent=self),
                     labels.MATERIALS:              EditMaterialsView(),
                     
                     labels.CANVAS:                 Graphics(),
                     labels.CRITICAL_POINT_REPORT_VIEW:  ResultCriticalPointView()}
            self._views = views
        
        self.toolbar = Toolbar()
        
        
        
        self.status_label = QLabel('status')
        self.progress = QProgressBar()
        self.status_label2 = QLabel()
        
        self.progress.setFixedWidth(200)  # Set the fixed width for the progress bar
        self.progress.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed) 
        
        
        spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        
        
        self.status_bar_layout = QHBoxLayout()

        self.status_bar_layout.addWidget(self.status_label)
        self.status_bar_layout.addSpacerItem(spacer)
        self.status_bar_layout.addWidget(self.status_label2)
        self.status_bar_layout.addWidget(self.progress)
        
        
        
        self.statusbar = QWidget()
        self.statusbar.setLayout(self.status_bar_layout)
        
        
        
        self.statusBar().addWidget(self.statusbar)
        self.statusBar().setVisible(True)
        
        
        self.addToolBar(self.toolbar)
        
        return self._views
                 
                  
            

        

    def put_views_in_tabs(self, layout):
        tab_widget = QTabWidget()
        for item in layout:
            tab_widget.addTab(self.views[item], item)
        return tab_widget
        
    def put_views_in_toolbox(self, layout):
        toolbox = QToolBox()
        tabs = {}
        for group_name, items in layout.items():
            if len(items) > 1:
                tabs[group_name] = self.put_views_in_tabs(items)
            else:
                tabs[group_name] = self.views[items[0]]
            toolbox.addItem(tabs[group_name], group_name)
        
        return toolbox, tabs
       
    
    def create_layout(self):
        toolbox, toolbox_tabs = self.put_views_in_toolbox(TOOLBOX_LAYOUT)
        self.toolbox = toolbox
        self.toolbox_tabs = toolbox_tabs
        
        self.result_container = self.put_views_in_tabs(RESULT_LAYOUT)
        
        self.main_container = QSplitter()
        
        # self.toolbox.setMinimumWidth(WIDGET_WIDTH+10)
        #self.toolbox.setMaximumWidth(WIDGET_WIDTH+20)
        
        
        self.main_container.addWidget(self.toolbox)
       
        self.main_container.addWidget(self.result_container)
        
        

        self.setCentralWidget(self.main_container)
        
        self.setContentsMargins(10, 10, 10, 10)
        
        self.main_container.setSizes([100, 500])

    
        
if __name__ == "__main__":
    def main():
        app = QApplication([])
        window = MainView()
        window.show()    
        app.exec_()
        return window
    
    
    window = main()

        



import yaml
import numpy as np


class Isotopes:
    def __init__(self):
        with open("../../isotopes.yml", "r") as f:
            dct = yaml.safe_load(f)
        
        self.isotopes = dct["Isotopes"]
        self.decay_chains = dct["Decay chains"]
        self.isotope_spectra = dct["Isotope spectra"]     
        
    def names(self):
        return [iso["Name"] for iso in self.isotopes]

    def get_spectrum(self, name):
        spectrum = list(self.isotope_spectra[name])
        for daughter, abundance in self.decay_chains.get(name, []):
            for energy, intensity in self.isotope_spectra[daughter]:
                spectrum.append([energy, intensity*abundance])
        return spectrum
    
    def get_SDEF(self, isotope):
        spectrum = np.array(self.get_spectrum(isotope))
        sorting = np.argsort(spectrum[:,0])

        energies = spectrum[sorting,0] * 1e-3
        intensities = spectrum[sorting,1]

        if len(energies) == 1:
            SDEF = f"{energies[0]} $ {isotope}"
        else:
            e_lines = np.array_split(energies, int(np.ceil(len(energies)/6)))
            i_lines = np.array_split(intensities, int(np.ceil(len(intensities)/6)))
            SDEF = f"d1 $ {isotope}\n"
            for prefix, lines in (("SI1 L  ", e_lines), ("SP1 D  ", i_lines)):
                for i, line in enumerate(lines):
                    if i == 0:
                        SDEF += prefix
                    else:
                        SDEF += " "*len(prefix)
                    SDEF += " ".join(["%9.5f"%x for x in line]) + "\n"
        return SDEF.rstrip()

class Materials:
    def __init__(self):
        with open("../../defaults.yml", "r") as f:
            dct = yaml.safe_load(f)            
        
        self.materials = [x for x in dct["Materials"] if x["Name"] != "None"]
        self.densities = {x["Name"]: x["Density"] for x in self.materials}
    
    def names(self):
        return [mat["Name"] for mat in self.materials]
    
    def get_MDEF(self, material):
        MDEF = {   
        ### WARNING: Don't just change these. Make sure the numbers 
        ### correspond to the correct materials in the templates
            "Water":           f'1  -{self.densities["Water"]}',
            "Lead":            f'2  -{self.densities["Lead"]}',
            "Concrete":        f'4  -{self.densities["Concrete"]}',
            "Concrete-Barite": f'5  -{self.densities["Concrete-Barite"]}',
            "Gypsum":          f'6  -{self.densities["Gypsum"]}',
            "Brick":           f'7  -{self.densities["Brick"]}',
            "Tungsten":        f'8  -{self.densities["Tungsten"]}',
        }[material]
        return MDEF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 10 11:35:41 2016

@author: Marcel
"""
from scipy.interpolate import CubicSpline
import numpy as np

def attenuation(isotope, material=None, thickness=None):
    unshielded = H10()
    
    
    
    
def ratio_H10_air_kerma(energy_keV):
    # Emperical function proposed by ICRU 47
    # 
    E0 = 9.85
    xi = np.log(energy_keV / E0)

    r = xi / (1.465 * xi**2 - 4.414 * xi + 4.789) \
        + 0.7006 * np.arctan(0.6519 * xi)
    return r

def H10(energy_keV, abundance = 1, add = True):
    """
    Calculates the h10 in uSv/h per MBq/m^2 for given photon energy or
    multiple energies.
    Args:
      energy_keV: photon energy(ies)
      abundance:  abundance for each photon energy
      add:        sum dose rates for all energies (default = True)
    Returns:
      dose rate in uSv/h per MBq/m^2
    """
    # convert tuple and list to numpy
    energy_keV = np.array(energy_keV)
    abundance = np.array(abundance)

    #ratio = np.interp(energy_keV, energies_keV, Hp10_ka)
    ratio = ratio_H10_air_kerma(energy_keV)

    h10 = ratio * kerma_air_rate(energy_keV, abundance, add = False)

    if add:
        h10 = np.sum(h10)
    return h10

def kerma_air_rate(energy_keV, abundance=1, add = True):
    """
    Calculates the air kerma in uGy/h per MBq/m^2 for given photon energy or
    multiple energies.
    Args:
      energy_keV: photon energy(ies)
      abundance:  abundance for each photon energy
      add:        sum dose rates for all energies (default = True)
    Returns:
      kerma in uGy/h per MBq/m^2
    """
    # air kerma : dKair/dt = A/(4*pi*l^2) * uk/p * E
    energy_keV = np.array(energy_keV, dtype=np.float64)
                          
    abundance = np.array(abundance, dtype=np.float64)

    # kerma rate at 1m for 1 Bq (A=1, l=1)
    Joule_per_eV = 1.60217662e-19

    energy_J = Joule_per_eV * energy_keV * 1000

    energy_absorption_coeff = linear_energy_absorption_coeff_air(energy_keV)

    # s^-1 --> h^-1 Gy--> uGy Bq --> MBq
    energy_absorption_coeff *= 3600 * 1e12

    # kerma in uGy/h per MBq/m^2
    kerma = abundance * energy_absorption_coeff * energy_J / (4 * np.pi)

    if add:
        kerma = np.sum(kerma)
    return kerma

def linear_energy_absorption_coeff_air(energy_keV):
    """
    Calculates the linear energy transfer coefficients by interpolation.
    Source data is obtained from the NIST XCOM database.
    https://physics.nist.gov/PhysRefData/XrayMassCoef/ComTab/air.html
    Args:
      energy_keV: photon energy(ies) in keV
    Returns:
      linear energy tranfer rate(s)
    """

    energy_keV = np.array(energy_keV)
    # source data (cut off at 15 keV)
    Energy_MeV =  [ 1.50E-02,
                    2.00E-02,
                    3.00E-02,
                    4.00E-02,
                    5.00E-02,
                    6.00E-02,
                    8.00E-02,
                    1.00E-01,
                    1.50E-01,
                    2.00E-01,
                    3.00E-01,
                    4.00E-01,
                    5.00E-01,
                    6.00E-01,
                    8.00E-01,
                    1.00E+00,
                    1.25E+00,
                    1.50E+00,
                    2.00E+00,
                    3.00E+00,
                    4.00E+00,
                    5.00E+00,
                    6.00E+00,
                    8.00E+00,
                    1.00E+01,
                    1.50E+01,
                    2.00E+01]

    u_en_p = [  1.3340E+00,
                5.3890E-01,
                1.5370E-01,
                6.8330E-02,
                4.0980E-02,
                3.0410E-02,
                2.4070E-02,
                2.3250E-02,
                2.4960E-02,
                2.6720E-02,
                2.8720E-02,
                2.9490E-02,
                2.9660E-02,
                2.9530E-02,
                2.8820E-02,
                2.7890E-02,
                2.6660E-02,
                2.5470E-02,
                2.3450E-02,
                2.0570E-02,
                1.8700E-02,
                1.7400E-02,
                1.6470E-02,
                1.5250E-02,
                1.4500E-02,
                1.3530E-02,
                1.3110E-02]

    
    interpolator = CubicSpline(Energy_MeV, u_en_p)
    
    #coeff = np.interp(np.log10(energy_keV/1e3), np.log10(Energy_MeV), u_en_p) # Units cm^2 per g
    coeff = interpolator(energy_keV/1e3)
    
    
    # plt.plot(Energy_MeV, u_en_p, 'd')
    # x = np.arange(0.015, 20, 0.001 )
    # plt.plot(x, interpolator(x), '-')
    
    
    coeff /= 10 # cm^2/g --> m^2/kg
    return coeff







import qtawesome as qta


from PyQt5.QtWidgets import QApplication

from PyQt5.QtWidgets import (QWidget, QCheckBox, QPushButton, 
                             QLineEdit, QSpinBox, QGridLayout, QLabel, 
                             QComboBox, QDoubleSpinBox, QColorDialog, 
                             QScrollBar, QMainWindow, QRadioButton)
                             
from PyQt5 import QtCore, QtGui
from PyQt5.QtCore import pyqtSignal


from pyrateshield.model import Model
from pyrateshield import labels
from pyrateshield.constants import CONSTANTS



MAX_VALUE = float('inf')
MAX_INT_VALUE = 99999999



def safe_get_value_from_widget(widget):
    if isinstance(widget, (EditValueWidget, QDoubleSpinBox, QSpinBox)):
        value = widget.value()
    elif isinstance(widget, (QCheckBox, QRadioButton)):
        value = widget.isChecked()
    elif isinstance(widget, QComboBox):
        value = widget.currentText()
    elif isinstance(widget, (QLineEdit, QLabel)):
         value = widget.text()
         
    else:
        raise TypeError(type(widget))
    return value

def safe_clear(widget):
    if isinstance(widget, (EditValueWidget, QDoubleSpinBox, QSpinBox, QLineEdit)):
        widget.clear()
    elif isinstance(widget, (QRadioButton, QCheckBox)):
        widget.setChecked(False)
    elif isinstance(widget, QComboBox):
        if widget.count() > 0:
            widget.setCurrentIndex(0)
        else:
            widget.setCurrentText('')
        
    elif isinstance(widget, EditValueWidget):
        widget.clear()


def safe_to_int(value):
    if value == '':
        value = 0
    else:
        value = int(round(float(value)))
    return value

def safe_to_float(value):
    if value == '':
        value = 0
    else:
        value = float(value)
    return value

def safe_set_value_to_widget(widget, value):
    # Set widget to a specified value
    # do not set when value equals current value
    # will not generate events when changed
    
    if isinstance(widget, QSpinBox):
        if widget.value() != safe_to_int(value):
            widget.setValue(safe_to_int(value))
    elif isinstance(widget, QDoubleSpinBox):
        if widget.value() != safe_to_float(value):
            widget.setValue(safe_to_float(value))
    elif isinstance(widget, (QLineEdit, QLabel)):
        if widget.text() != str(value):
            widget.setText(str(value))
    elif isinstance(widget, (QRadioButton, QCheckBox)):
        if widget.isChecked() != value:
            widget.setChecked(value)
    elif isinstance(widget, QScrollBar):
        if widget.value() != safe_to_int(value):
            widget.setValue(safe_to_int(value))
    elif isinstance(widget, QComboBox):
        if value is None:
            value = ''
        if isinstance(value, str):
            value = widget.findText(str(value))        
        if widget.currentIndex != safe_to_int(value):
            widget.setCurrentIndex(safe_to_int(value))
    elif isinstance(widget, EditValueWidget):
        widget.setValue(value)
            
    else:
        raise TypeError(f'Unsupported widget type {type(widget)}')
        
class BlockValueChangedSignal:
    def __init__(self, view=None):
        self.view = view
        
    def __enter__(self):
        self.old_value = self.view.valueChangedIsBlocked()
        self.view.setValueChangedBlocked(True)
        
    def __exit__(self, *args):
        self.view.setValueChangedBlocked(self.old_value)
        
class QDoubleSpinBoxInfinite(QDoubleSpinBox):

    def __init__(self, *args, **kwargs):
        super(QDoubleSpinBox, self).__init__(*args, **kwargs)

        self.setMinimum(float('-inf'))
        self.setMaximum(float('inf'))

    def keyPressEvent(self, e: QtGui.QKeyEvent):
        if e.key() == QtCore.Qt.Key_End:
            self.setValue(self.maximum())
        elif e.key() == QtCore.Qt.Key_Home:
            self.setValue(self.minimum())
        else:
            super(QDoubleSpinBox, self).keyPressEvent(e)

class EditValueWidget(QWidget):
    valueChanged = pyqtSignal(object)
    
    def __init__(self, parent=None):

        QWidget.__init__(self, parent=parent)
        self.layout = QGridLayout()
        self.create_widgets()
        self.create_layout()
        self.set_callbacks()
        self.setLayout(self.layout)
    
    def create_widgets(self):
        pass
    
    def set_callbacks(self):
        pass
    
    def value_changed(self):
        if not self.parent().valueChangedIsBlocked():
            self.valueChanged.emit(self.value())
        
    def create_layout(self):
        pass
    
    def value(self):
        pass
    
    def setValue(self):
        pass
   
    
class PositionWidget(EditValueWidget):
    _position_x_text = "X [{unit}]:"
    _position_y_text = "Y [{unit}]:"
    
    def __init__(self, *args, unit='cm', header=labels.POSITION, **kwargs):
        self.unit = unit
        self.header = header
        super().__init__(*args, **kwargs)
    
    def set_callbacks(self):          
        self.x.valueChanged.connect(self.value_changed)
        self.y.valueChanged.connect(self.value_changed)
    
    
    def create_widgets(self):
        x = QDoubleSpinBox(decimals=1)
        x.setRange(-MAX_VALUE, MAX_VALUE)
    
        y = QDoubleSpinBox(decimals=1)
        y.setRange(-MAX_VALUE, MAX_VALUE)
        
        self.x = x
        self.y = y
        self.position_label = QLabel(self.header)
        self.position_label.setStyleSheet('font-weight: bold')
        self.position_x_label = QLabel(self._position_x_text.format(unit=self.unit))
        self.position_y_label = QLabel(self._position_y_text.format(unit=self.unit))
        
    def setValue(self, value):
        safe_set_value_to_widget(self.x, value[0])
        safe_set_value_to_widget(self.y, value[1])
        
    def value(self):
        return [self.x.value(), self.y.value()]
    
    def create_layout(self):
        self.layout.addWidget(self.position_label, 0, 0, 1, 2)

        self.layout.addWidget(self.position_x_label, 1, 0)
        self.layout.addWidget(self.x, 1, 1)
     
        self.layout.addWidget(self.position_y_label, 2, 0)
        self.layout.addWidget(self.y, 2, 1)
        # row += 1
        # self.layout.addWidget(self.position_button, row, 0, 1, 2)
        
    def clear(self):
        safe_clear(self.x)
        safe_clear(self.y)


class VerticesWidget(EditValueWidget):
   
    
    def __init__(self, *args, unit='cm', **kwargs):
        self.unit = unit
        super().__init__(*args, **kwargs)
        
    def set_callbacks(self):          
        self.x1.valueChanged.connect(self.value_changed)
        self.y1.valueChanged.connect(self.value_changed)
        self.x2.valueChanged.connect(self.value_changed)
        self.y2.valueChanged.connect(self.value_changed)
    
        
    def create_widgets(self):
        self.x1 = QDoubleSpinBox(decimals=1)
        self.x1.setRange(-MAX_VALUE, MAX_VALUE)
    
        self.y1 = QDoubleSpinBox(decimals=1)
        self.y1.setRange(-MAX_VALUE, MAX_VALUE)
        
        self.x2 = QDoubleSpinBox(decimals=1)
        self.x2.setRange(-MAX_VALUE, MAX_VALUE)
    
        self.y2 = QDoubleSpinBox(decimals=1)
        self.y2.setRange(-MAX_VALUE, MAX_VALUE)
        
        
        self.label = QLabel(labels.VERTICES)
        self.label.setStyleSheet('font-weight: bold')
        
        position_x1_text = f'X1 [{self.unit}]:'
        position_y1_text = f'Y1 [{self.unit}]:'
        position_x2_text = f'X2 [{self.unit}]:'
        position_y2_text = f'Y2 [{self.unit}]:'
        
        self.position_x1_label = QLabel(position_x1_text)
        self.position_y1_label = QLabel(position_y1_text)
        self.position_x2_label = QLabel(position_x2_text)
        self.position_y2_label = QLabel(position_y2_text)
        
    def setValue(self, value):
   
        safe_set_value_to_widget(self.x1, value[0][0])
        safe_set_value_to_widget(self.y1, value[0][1])
        safe_set_value_to_widget(self.x2, value[1][0])
        safe_set_value_to_widget(self.y2, value[1][1])
        
    def value(self):
        vertices = [[self.x1.value(), self.y1.value()],
                    [self.x2.value(), self.y2.value()]]
       
        return vertices
    
    def create_layout(self):
        self.layout.addWidget(self.label, 0, 0, 1, 2)

        self.layout.addWidget(self.position_x1_label, 1, 0)
        self.layout.addWidget(self.x1, 1, 1)
     
        self.layout.addWidget(self.position_y1_label, 2, 0)
        self.layout.addWidget(self.y1, 2, 1)
        
        self.layout.addWidget(self.position_x2_label, 3, 0)
        self.layout.addWidget(self.x2, 3, 1)
     
        self.layout.addWidget(self.position_y2_label, 4, 0)
        self.layout.addWidget(self.y2, 4, 1)
        # row += 1
        # self.layout.addWidget(self.position_button, row, 0, 1, 2)
        
    def clear(self):
        for widget in self.x1, self.x2, self.y1, self.y2:
            safe_clear(widget)
        
     
class IsotopeWidget(EditValueWidget):
  
    def create_layout(self):        
        self.layout.addWidget(self.isotope_label, 0, 0)
        self.layout.addWidget(self.isotope_input, 0, 1)
        self.layout.addWidget(self.halflife_label, 1, 0)
        self.layout.addWidget(self.halflife_value, 1, 1)
    
        
    def create_widgets(self):
        self.isotope_input = QComboBox()
        self.isotope_input.addItems(CONSTANTS.get_isotope_name_list())
        self.isotope_label = QLabel(labels.ISOTOPE)
        self.halflife_label = QLabel(labels.HALF_LIFE)
        self.halflife_value = QLabel('empty')
        
    def set_callbacks(self):
        callback = self.value_changed
        self.isotope_input.currentTextChanged.connect(callback)
        
    def update_halflife_label(self):
        isotope = CONSTANTS.get_isotope_by_name(self.value())
     
        safe_set_value_to_widget(self.halflife_value, isotope.half_life)
    
    def value_changed(self):
        self.update_halflife_label()
        
        self.valueChanged.emit(self.value())
        
    def value(self):
        return self.isotope_input.currentText()
    
    def setValue(self, value):
        self.isotope_input.setCurrentText(value) 
        self.update_halflife_label()
        
    def clear(self):
        for widget in self.isotope_input, self.halflife_value:
            safe_clear(widget)
        



class SelfShieldingWidget(EditValueWidget):
    def create_widgets(self):    
        self.self_shielding_list = QComboBox()
        self.self_shielding_list.addItems(CONSTANTS.self_shielding_options)
        self.self_shielding_label = QLabel(labels.SELF_SHIELDING)
        self.self_shielding_factor_label = QLabel(labels.SELF_SHIELDING_FACTOR)
        self.self_shielding_factor_input = QDoubleSpinBox()
        self.self_shielding_factor_input.setSingleStep(0.01)
        self.self_shielding_factor_input.setRange(0, MAX_VALUE)
        self.self_shielding_factor_input.setDecimals(3)
        self.self_shielding_factor_input.setValue(1.0)     
    
    def create_layout(self):
        self.layout = QGridLayout()
        self.layout.addWidget(self.self_shielding_label, 0, 0)
        self.layout.addWidget(self.self_shielding_list, 0, 1)
        self.layout.addWidget(self.self_shielding_factor_label, 1, 0)
        self.layout.addWidget(self.self_shielding_factor_input, 1, 1)
        
    def set_callbacks(self):
       
        callback = self.value_changed
        
        self.self_shielding_list.currentTextChanged.connect(callback)        
        self.self_shielding_factor_input.valueChanged.connect(callback)
        
    def value_changed(self):
        self.valueChanged.emit(self.value())
        if isinstance(self.value(), str):
            enabled = False
        else:
            enabled = True
            
        self.self_shielding_factor_label.setEnabled(enabled)
        self.self_shielding_factor_input.setEnabled(enabled)
        
        
    def value(self):
        list_value = self.self_shielding_list.currentText()
        if list_value == labels.SELF_SHIELDING_FACTOR:
            return self.self_shielding_factor_input.value()
        else:
            return list_value
        
    def setValue(self, value):
        if not isinstance(value, str):
            self.self_shielding_list.setCurrentText(labels.SELF_SHIELDING_FACTOR)
        else:
            self.self_shielding_list.setCurrentText(value)
            
    def clear(self):
        for widget in self.self_shielding_factor_input, self.self_shielding_list:
            safe_clear(widget)
            
class MaterialsWidget(EditValueWidget):
    def create_widgets(self):
        self.material1_label = QLabel(labels.MATERIAL + ' 1')
        self.material1_list = QComboBox()
    
        self.thickness1_label = QLabel(labels.THICKNESS)
        self.thickness1_input = QDoubleSpinBox(decimals=3)
    
        self.material2_label = QLabel(labels.MATERIAL + ' 2')
        self.material2_list = QComboBox()
        
        self.thickness2_label = QLabel(labels.THICKNESS)
        self.thickness2_input = QDoubleSpinBox(decimals=3)
    
    def create_layout(self):
        row = 0
        self.layout.addWidget(self.material1_label, row, 0)
        self.layout.addWidget(self.material1_list, row, 1)
        row += 1
        self.layout.addWidget(self.thickness1_label, row, 0)
        self.layout.addWidget(self.thickness1_input, row, 1)
        row += 1
        self.layout.addWidget(self.material2_label, row, 0)
        self.layout.addWidget(self.material2_list, row, 1)
        row += 1
        self.layout.addWidget(self.thickness2_label, row, 0)
        self.layout.addWidget(self.thickness2_input, row, 1)
        
    def set_callbacks(self):
        
        callback = self.value_changed
        
        self.material1_list.currentTextChanged.connect(callback)
        self.material2_list.currentTextChanged.connect(callback)
        self.thickness1_input.valueChanged.connect(callback)
        self.thickness2_input.valueChanged.connect(callback)
    
    def value(self):
        materials1 = safe_get_value_from_widget(self.material1_list)
        thickness1 = safe_get_value_from_widget(self.thickness1_input)
        thickness2 = safe_get_value_from_widget(self.thickness2_input)
        materials2 = safe_get_value_from_widget(self.material2_list)
        
        # if thickness1 == 0:
        #     materials1 = None
        # if thickness2 == 0:
        #     materials2 = None
                    
        return [[materials1, thickness1], [materials2, thickness2]]
    
    def setValue(self, value):

        materials1 = value[0][0]
        thickness1 = value[0][1]
        
        if len(value) == 2:
            materials2 = value[1][0]
            thickness2 = value[1][1]
        else:
            materials2 = labels.EMPTY_MATERIAL
            thickness2 = 0
            
             
        safe_set_value_to_widget(self.material1_list, materials1)
        safe_set_value_to_widget(self.thickness1_input, thickness1)
        
        
        
        if len(value) == 2:
            
            safe_set_value_to_widget(self.material2_list, materials2)
            safe_set_value_to_widget(self.thickness2_input, thickness2 )
        else:
            safe_set_value_to_widget(self.material2_list, labels.EMPTY_MATERIAL)
            safe_set_value_to_widget(self.thickness2_input, 0)
            
    def clear(self):
        for widget in (self.material1_list, self. material2_list,
                       self.thickness1_input, self.thickness2_input):
            safe_clear(widget)
    
            
class ModelItemsWidget(EditValueWidget):
    _name_text = "Name:"
    _value_widgets = None
    enabled = None
    name_input = None
    list = None
    valueChanged = pyqtSignal(str, object)
    
    _signal_blocked = False
    
    def __init__(self, *args, parent=None, **kwargs):
        super().__init__(*args, parent=parent, **kwargs)
        self.layout.setRowStretch(self.layout.rowCount(), 1)
    
    def setValueChangedBlocked(self, block=True):
        self._signal_blocked = block
        
    def valueChangedIsBlocked(self):
        return self._signal_blocked
    
    def add_list(self):
        self.list = QComboBox()
        row = self.layout.rowCount()
        self.layout.addWidget(self.list, row, 0, 1, 2)
        
    def add_enabled(self):
        self.enabled = QCheckBox('Enabled')
        row = self.layout.rowCount()
        self.layout.addWidget(self.enabled, row, 0, 1, 2)
    
    def add_name(self):
        self.name_label = QLabel(self._name_text)
        self.name_input = QLineEdit()
        
        row = self.layout.rowCount()
        
        self.layout.addWidget(self.name_label, row, 0)
        self.layout.addWidget(self.name_input, row, 1)
        
    def value_widgets(self):
        if self._value_widgets is None:
            self._value_widgets = {}
            if self.enabled is not None:
                self._value_widgets[labels.ENABLED] = self.enabled
            if self.name_input is not None:
                self._value_widgets[labels.NAME] = self.name_input
        return self._value_widgets
    
    def clear(self):
        with BlockValueChangedSignal(self):
            for widget in self.value_widgets().values():
                safe_clear(widget)
            if hasattr(self, 'list'):
                safe_clear(self.list)
           
    
    def value_changed(self, label): 
        if not self._signal_blocked:
            self.valueChanged.emit(label, self.value(label))
     
            
    def value(self, label):
       
        widget = self.value_widgets()[label]
        value = safe_get_value_from_widget(widget)
        return value
    
    def setValue(self, label, value):
        widget = self.value_widgets()[label]
        
        safe_set_value_to_widget(widget, value)
        
        if label == labels.NAME:            
            safe_set_value_to_widget(self.name_input, value)
            
        if label == labels.ENABLED and value is not None:
            self.setEnabled(value)
            
        
            
    def values(self):
        return {label: safe_get_value_from_widget(widget)\
                for label, widget in self.value_widgets().items()}
            
    def setValues(self, values):     
        for label, value in values.items():
            self.setValue(label, value)
            
        
    def set_callbacks(self):
        if self.enabled is not None:
            callback = lambda _, label=labels.ENABLED: self.value_changed(label)
            self.enabled.stateChanged.connect(callback)
        if self.name_input is not None:
            callback = lambda label=labels.NAME: self.value_changed(label)
            self.name_input.returnPressed.connect(callback)
            
   

    def setEnabled(self, enabled):
        widgets = [item for item in self.children()\
                   if isinstance(item, QWidget)]
        
        if self.list in widgets: widgets.remove(self.list)
        if self.enabled in widgets: widgets.remove(self.enabled)
        
        for widget in widgets:
            widget.setEnabled(enabled)
        
            

class EditSourcesNMView(ModelItemsWidget):
    LABEL = labels.SOURCES_NM
    explanation = ("To add a new source select Source NM from the toolbar and "
                   "click on the floorplan to add a source at that position."
                   "\n\n"
                   "When setting self shielding to 'Body', pyshield will "
                   "assume 10 cm of water as additional shielding (buildup "
                   "and attenuation). Radtracer will use a pre-simulated "
                   "spectrum after additional 10 cm of "
                   "shielding with water. Note that large differences in results "
                   "may occur between pyshield and radtracer when setting self "
                   "shielding to 'Body.'\n\n"
                   "Note that setting the self shielding to 'Body' may increase "
                   "the dose rates for pyshield for some isotopes. This is due "
                   "to the buildup in 10 cm of water. Using a user defined fixed "
                   "factor could provide a better results for pyshield.")
    
    
    
                
    def create_widgets(self):
        
        super().create_widgets()
        self.position = PositionWidget(parent=self)
        
        self.isotope = IsotopeWidget(parent=self)
        
        self.duration = QDoubleSpinBox(self, decimals=3)
        self.duration.setRange(0, MAX_VALUE)
        self.duration_label = QLabel(labels.DURATION)
        
               
        self.activity = QDoubleSpinBox(self, decimals=2)
        self.activity.setRange(0, MAX_VALUE)
        self.activity_label = QLabel(labels.ACTIVITY)

        self.number_of_exams_label = QLabel(labels.NUMBER_OF_EXAMS, parent=self)
        self.number_of_exams_input = QSpinBox(self)
        self.number_of_exams_input.setRange(0, MAX_INT_VALUE)
        

        self.decay_correction = QCheckBox(labels.APPLY_DECAY_CORRECTION,
                                          parent=self)
        
        self.self_shielding = SelfShieldingWidget(parent=self)
        
        self.clearance_label = QLabel('Clearance Model', parent=self)
        self.clearance_list = QComboBox(self)
        
    def value_widgets(self):
       if self._value_widgets is None:
           self._value_widgets = super().value_widgets()
           self._value_widgets[labels.POSITION] = self.position
           self._value_widgets[labels.DURATION] = self.duration
           self._value_widgets[labels.ACTIVITY] = self.activity
           self._value_widgets[labels.ISOTOPE] = self.isotope
           self._value_widgets[labels.SELF_SHIELDING] = self.self_shielding
           self._value_widgets[labels.NUMBER_OF_EXAMS] = self.number_of_exams_input
           self._value_widgets[labels.APPLY_DECAY_CORRECTION] = self.decay_correction
           self._value_widgets[labels.CLEARANCE] = self.clearance_list
       return self._value_widgets
    
    
            

    def create_layout(self):
        
        self.add_list()
        self.add_name()

        row = self.layout.rowCount()
        
        self.layout.addWidget(self.position, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.isotope, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.decay_correction, row, 0, 1, 2)
        
        
        
        row += 1
        
        self.layout.addWidget(self.number_of_exams_label, row, 0)
        self.layout.addWidget(self.number_of_exams_input, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.activity_label, row, 0)
        self.layout.addWidget(self.activity, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.duration_label, row, 0)
        self.layout.addWidget(self.duration, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.self_shielding, row, 0, 1, 2)
        
        
        row += 1

        self.layout.addWidget(self.clearance_label, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.clearance_list, row, 0, 1, 2)
        
        
           
        self.add_enabled()
           
    
    
    def set_callbacks(self):
        super().set_callbacks()
        
        
        label = labels.POSITION
        callback = lambda _, label=label: self.value_changed(label)        
        self.position.valueChanged.connect(callback)
        
        label = labels.ISOTOPE
        callback = lambda _, label=label: self.value_changed(label)        
        self.isotope.valueChanged.connect(callback)
        
        label = labels.APPLY_DECAY_CORRECTION
        callback = lambda _, label=label: self.value_changed(label)        
        self.decay_correction.stateChanged.connect(callback)
        
        label = labels.NUMBER_OF_EXAMS
        callback = lambda _, label=label: self.value_changed(label)
        self.number_of_exams_input.valueChanged.connect(callback)
        
        label = labels.ACTIVITY
        callback = lambda _, label=label: self.value_changed(label)
        self.activity.valueChanged.connect(callback)
        
        label = labels.DURATION
        callback = lambda _, label=label: self.value_changed(label)        
        self.duration.valueChanged.connect(callback)
        
        label = labels.CLEARANCE
        callback = lambda _, label=label: self.value_changed(label)        
        self.clearance_list.currentIndexChanged.connect(callback)
        
        label = labels.CLEARANCE
        callback = lambda _, label=label: self.value_changed(label)        
        self.clearance_list.currentIndexChanged.connect(callback)
        
    
         
class EditClearancesView(ModelItemsWidget):
    explanation = ("\nUp to two biological fraction can be defined with a"
                   " corresponding biological halflife. For each fraction "
                   "pyrateshield will calculate an effective halflife by "
                   "combining the physical halflife and biological halflife. "
                   "If (physical) decay is not checked in the Sources NM tab "
                   "physical decay will be ignored and only biological "
                   "decay will be applied.\n\n"
                   "If the fractions add up to less than 1, no biological"
                   "decay correction will be applied to the remaining fraction"
                   "\n\n"
                   "Optionally fractions can be split in time. The first "
                   "fraction will be integrated until the split time. The "
                   "second fraction and if applicable the remaining fraction "
                   "will be integrated from the split time. Integration will "
                   "always stop after source duration.")
    
    def create_widgets(self):
        super().create_widgets()
        self.add_list()
        self.add_name()
        
        self.decay_fraction1_checkbox = QCheckBox(labels.DECAY_FRACTION1)
        self.decay_fraction1 = QDoubleSpinBox(decimals=2)
        self.decay_fraction1.setRange(0, 1)
        self.decay_fraction1.setSingleStep(0.05)
        

        self.decay_fraction2_checkbox = QCheckBox(labels.DECAY_FRACTION2)
        self.decay_fraction2 = QDoubleSpinBox(decimals=2)
        self.decay_fraction2.setRange(0, 1)
        self.decay_fraction2.setSingleStep(0.05)
        
        # Prevent halflives to become zero
        self.half_life_label1 = QLabel(labels.HALF_LIFE)
        self.half_life_value1 = QDoubleSpinBoxInfinite(decimals=2)
        self.half_life_value1.setRange(0.01, MAX_VALUE)
        
        self.half_life_label2 = QLabel(labels.HALF_LIFE)
        self.half_life_value2 = QDoubleSpinBoxInfinite(decimals=2)
        self.half_life_value2.setRange(0.01, MAX_VALUE)
        
        self.split_time_checkbox = QCheckBox(labels.ENABLE_SPLIT_FRACTIONS)
        self.split_time_label = QLabel(labels.SPLIT_FRACTION_TIME)
        self.split_time_input = QDoubleSpinBox(decimals=1)
        self.split_time_input.setRange(0, MAX_VALUE)
        
        
        
        self.new_button = QPushButton('Add', self)
        self.delete_button = QPushButton("Delete")
        
    def value_widgets(self):
        if self._value_widgets is None:
            self._value_widgets = super().value_widgets()
            self._value_widgets[labels.APPLY_FRACTION1] = self.decay_fraction1_checkbox
            self._value_widgets[labels.APPLY_FRACTION2] = self.decay_fraction2_checkbox
            self._value_widgets[labels.DECAY_FRACTION1] = self.decay_fraction1
            self._value_widgets[labels.DECAY_FRACTION2] = self.decay_fraction2
            self._value_widgets[labels.HALFLIFE1] = self.half_life_value1
            self._value_widgets[labels.HALFLIFE2] = self.half_life_value2
            self._value_widgets[labels.ENABLE_SPLIT_FRACTIONS] = self.split_time_checkbox
            self._value_widgets[labels.SPLIT_FRACTION_TIME] = self.split_time_input
        return self._value_widgets
            
    def value_changed(self, label):
        super().value_changed(label)
        self.update_available_widgets()
    
    def setValues(self, values):        
        super().setValues(values)
        values = values or {}
        if labels.NAME in values.keys()\
            and values[labels.NAME] == labels.EMPTY_CLEARANCE:            
            self.setEnabled(False)
        else:                  
            self.setEnabled(True)            
            self.update_available_widgets()

    def update_available_widgets(self, _=None):
      
        enabled = self.name_input.isEnabled()
        if not enabled: return
        self.decay_fraction1_checkbox.setChecked(True)
        self.decay_fraction1_checkbox.setEnabled(False)
        
        self.decay_fraction1.setEnabled(True)
        self.half_life_value1.setEnabled(True)
        self.half_life_label1.setEnabled(True)
        
        
        enabled = self.decay_fraction2_checkbox.isChecked()
        
        self.decay_fraction2.setEnabled(enabled)
        self.half_life_value2.setEnabled(enabled)
        self.half_life_label2.setEnabled(enabled)
        
        
        enabled = self.split_time_checkbox.isChecked()
        self.split_time_input.setEnabled(enabled)
        self.split_time_label.setEnabled(enabled)
        
        if enabled:
            self.decay_fraction2_checkbox.setChecked(True)
            self.decay_fraction2_checkbox.setEnabled(False)
            self.decay_fraction2.setEnabled(False)
            self.half_life_value2.setEnabled(True)
        else:
            self.decay_fraction2_checkbox.setEnabled(True)

    def set_callbacks(self):
        super().set_callbacks()
        label = labels.HALFLIFE1
        callback = lambda _, label=label: self.value_changed(label)        
        self.half_life_value1.valueChanged.connect(callback)
        
        label = labels.HALFLIFE2
        callback = lambda _, label=label: self.value_changed(label)        
        self.half_life_value2.valueChanged.connect(callback)

        label = labels.SPLIT_FRACTION_TIME
        callback = lambda _, label=label: self.value_changed(label)        
        self.split_time_input.valueChanged.connect(callback)
        
        label = labels.DECAY_FRACTION1
        callback = lambda _, label=label: self.value_changed(label)        
        self.decay_fraction1.valueChanged.connect(callback)
        

        label = labels.DECAY_FRACTION2
        callback = lambda _, label=label: self.value_changed(label)        
        self.decay_fraction2.valueChanged.connect(callback)
        
        label = labels.APPLY_FRACTION1
        callback = lambda _, label=label: self.value_changed(label)        
        self.decay_fraction1_checkbox.stateChanged.connect(callback)
        
        label = labels.APPLY_FRACTION2
        callback = lambda _, label=label: self.value_changed(label)        
        self.decay_fraction2_checkbox.stateChanged.connect(callback)
        
        label = labels.ENABLE_SPLIT_FRACTIONS
        callback = lambda _, label=label: self.value_changed(label)        
        self.split_time_checkbox.stateChanged.connect(callback)
        
        self.list.currentIndexChanged.connect(self.update_available_widgets)
    
    def create_layout(self):
        super().create_layout()
        
        row = self.layout.rowCount() + 1
     
        self.layout.addWidget(self.decay_fraction1_checkbox, row, 0)
        self.layout.addWidget(self.decay_fraction1, row, 1)
        row += 1
        self.layout.addWidget(self.half_life_label1, row, 0)
        self.layout.addWidget(self.half_life_value1, row, 1)    
                
        row += 1
        
        self.layout.addWidget(self.decay_fraction2_checkbox, row, 0)
        self.layout.addWidget(self.decay_fraction2, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.half_life_label2, row, 0)
        self.layout.addWidget(self.half_life_value2, row, 1)  
        
       
        row += 1
        
        self.layout.addWidget(self.split_time_checkbox, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.split_time_label, row, 0)
        self.layout.addWidget(self.split_time_input, row, 1)
   
        row += 1
        
        self.layout.addWidget(self.new_button, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.delete_button, row, 0, 1, 2)
    
    def setEnabled(self, enabled):
        super().setEnabled(enabled)
        self.new_button.setEnabled(True)
        



         

class EditSourcesXrayView(ModelItemsWidget):
 
    explanation = ("To add a new source select Source Xray from the toolbar and "
                   "click on the floorplan to add a source at that position.")
                   
    
    @property
    def kvp_labels(self):
        return [str(item.kvp) for item in CONSTANTS.xray]
    
    def create_widgets(self):
        super().create_widgets()
        self.add_list()
        self.add_name()
        
        self.position = PositionWidget()
        
        self.number_of_exams_input = QSpinBox()
        self.number_of_exams_input.setRange(0, MAX_INT_VALUE)
        self.number_of_exams_label = QLabel(labels.NUMBER_OF_EXAMS)
        
        self.kvp = QComboBox()
        self.kvp.addItems(self.kvp_labels) 
        self.kvp_label = QLabel(labels.KVP)

        self.dap = QDoubleSpinBox()
        self.dap.setRange(0, MAX_VALUE)
        self.dap_label = QLabel(labels.DAP)
        
    def value_widgets(self):
        if self._value_widgets is None:
            self._value_widgets = super().value_widgets()
            self._value_widgets[labels.POSITION] = self.position
            self._value_widgets[labels.NUMBER_OF_EXAMS] = self.number_of_exams_input
            self._value_widgets[labels.KVP] = self.kvp
            self._value_widgets[labels.DAP] = self.dap
        return self._value_widgets
        
    def set_callbacks(self):
        super().set_callbacks()
        label = labels.POSITION
        callback = lambda _, label=label: self.value_changed(label)        
        self.position.valueChanged.connect(callback)
        
        
        label = labels.NUMBER_OF_EXAMS
        callback = lambda _, label=label: self.value_changed(label)        
        self.number_of_exams_input.valueChanged.connect(callback)
        
        label = labels.KVP
        callback = lambda _, label=label: self.value_changed(label)        
        self.kvp.currentTextChanged.connect(callback)
        
        label = labels.DAP
        callback = lambda _, label=label: self.value_changed(label)        
        self.dap.valueChanged.connect(callback)

    def create_layout(self):
        
        row = self.layout.rowCount() + 1
        self.layout.addWidget(self.number_of_exams_label, row, 0)
        self.layout.addWidget(self.number_of_exams_input, row, 1)
        
        row += 1
        self.layout.addWidget(self.kvp_label, row, 0)
        self.layout.addWidget(self.kvp, row, 1)
        row += 1
        self.layout.addWidget(self.dap_label, row, 0)
        self.layout.addWidget(self.dap, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.position, row, 0, 1, 2)
        
        self.add_enabled()
        



        
class EditSourcesCTView(ModelItemsWidget):
  
    explanation = ("To add a new source select Source CT from the toolbar and "
                   "click on the floorplan to add a source at that position.")
   
    
    @property
    def kvp_labels(self):
        return [str(item.kvp) for item in CONSTANTS.ct]
    
    def create_widgets(self):
        super().create_widgets()
        self.add_list()
        self.add_name()
        self.position = PositionWidget()
        
        self.number_of_exams_input = QSpinBox()
        self.number_of_exams_input.setRange(0, MAX_INT_VALUE)
        self.number_of_exams_label = QLabel(labels.NUMBER_OF_EXAMS)

        self.kvp = QComboBox()
        self.kvp.addItems(self.kvp_labels) 
        self.kvp_label = QLabel(labels.KVP)

        self.body_part = QComboBox()
        self.body_part.addItems(CONSTANTS.CT_body_part_options ) 
        self.body_part_label = QLabel(labels.CT_BODY_PART)

        self.dlp = QDoubleSpinBox()
        self.dlp.setRange(0, MAX_VALUE)
        self.dlp_label = QLabel(labels.DLP)
        
    def value_widgets(self):
        if self._value_widgets is None:
            self._value_widgets = super().value_widgets()
            self._value_widgets[labels.POSITION] = self.position
            self._value_widgets[labels.NUMBER_OF_EXAMS] = self.number_of_exams_input
            self._value_widgets[labels.KVP] = self.kvp
            self._value_widgets[labels.CT_BODY_PART] = self.body_part
            self._value_widgets[labels.DLP] = self.dlp
        return self._value_widgets

        
    def set_callbacks(self):
        super().set_callbacks()
        
        label = labels.POSITION
        callback = lambda _, label=label: self.value_changed(label)        
        self.position.valueChanged.connect(callback)
        
        
        label = labels.NUMBER_OF_EXAMS
        callback = lambda _, label=label: self.value_changed(label)                
        self.number_of_exams_input.valueChanged.connect(callback)
        
        label = labels.KVP
        callback = lambda _, label=label: self.value_changed(label)        
        self.kvp.currentTextChanged.connect(callback)
 
        
        label = labels.DLP
        callback = lambda _, label=label: self.value_changed(label)        
        self.dlp.valueChanged.connect(callback)
                
        label = labels.CT_BODY_PART
        callback = lambda _, label=label: self.value_changed(label)        
        self.body_part.currentTextChanged.connect(callback)
        
        
    def create_layout(self):
        row = self.layout.rowCount() + 1
        self.layout.addWidget(self.number_of_exams_label, row, 0)
        self.layout.addWidget(self.number_of_exams_input, row, 1)
        row += 1
        self.layout.addWidget(self.kvp_label, row, 0)
        self.layout.addWidget(self.kvp, row, 1)
        row += 1
        self.layout.addWidget(self.body_part_label, row, 0)
        self.layout.addWidget(self.body_part, row, 1)
        row += 1
        self.layout.addWidget(self.dlp_label, row, 0)
        self.layout.addWidget(self.dlp, row, 1)
        row += 1 
        self.layout.addWidget(self.position, row, 0, 1, 2)
        
        self.add_enabled()

class EditCriticalPointsView(ModelItemsWidget):
    LABEL = labels.CRITICAL_POINTS
    explanation = ("To add a new critcal point select Critical Point from "
                   "the toolbar and click on the floorplan to add a "
                   "critical point at that position.")
              
    
     

    def create_widgets(self):
        self.add_list()
        self.add_name()
        super().create_widgets()
        
        self.position = PositionWidget()
        self.occupancy_factor_label = QLabel("Occupancy Factor:")
        self.occupancy_factor_input = QDoubleSpinBox()
        self.occupancy_factor_input.setSingleStep(0.05)
        self.occupancy_factor_input.setRange(0, 1)
        self.occupancy_factor_input.setValue(1)
    
    def value_widgets(self):
        if self._value_widgets is None:
            self._value_widgets = super().value_widgets()
            self._value_widgets[labels.OCCUPANCY_FACTOR] = self.occupancy_factor_input
            self._value_widgets[labels.POSITION] = self.position
        return self._value_widgets

    def set_callbacks(self):
        super().set_callbacks()
        label = labels.OCCUPANCY_FACTOR
        callback = lambda _, label=label: self.value_changed(label)
    
        self.occupancy_factor_input.valueChanged.connect(callback)
        
        label = labels.POSITION
        callback = lambda _, label=label: self.value_changed(label)
    
        self.position.valueChanged.connect(callback)
        
  
        
    def create_layout(self):
        super().create_layout()

        row = self.layout.rowCount() + 1
        
        self.layout.addWidget(self.position, row, 0, 1, 2)

        row += 1
        
        self.layout.addWidget(self.occupancy_factor_label, row, 0)
        self.layout.addWidget(self.occupancy_factor_input, row, 1)
        
        self.add_enabled()
    
        
class EditShieldingsView(ModelItemsWidget):
    
    _color = 'r'
    _DEFAULT_THICKNESS = 1
    _DEFAULT_LINEWIDTH = 1
    
   
    def create_widgets(self):
        super().create_widgets()
        self.add_list()
        self.add_name()
        #icon = qta.icon('fa5s.circle', color='red')
        self.color_button = QPushButton("Select Color", self)
        #self.color_button.setIcon(icon)

        
        self.materials = MaterialsWidget()
        # materials = [material.name for material in CONSTANTS.materials]
        
        # self.material1_list.addItems(materials)
        # self.material2_list.addItems(materials)

       
        
        self.line_width_label = QLabel(labels.LINEWIDTH)
        self.line_width = QDoubleSpinBox()
        
        
        self.new_button = QPushButton('Add', self)
        self.delete_button = QPushButton("Delete")
        

        
    def value_widgets(self):
        if self._value_widgets is None:
            self._value_widgets = super().value_widgets()
            self._value_widgets[labels.MATERIALS] = self.materials
            self._value_widgets[labels.LINEWIDTH] = self.line_width
        return self._value_widgets
    
    def value(self, label):
        if label == labels.COLOR:
            return self._color
        else:
            return super().value(label)
    
    def setValues(self, values):
        super().setValues(values)
        values = values or {}
        if labels.NAME in values.keys()\
            and values[labels.NAME] == labels.EMPTY_SHIELDING:        
            self.setEnabled(False)
        else:              
            self.setEnabled(True)

        
    def values(self, label):
        values = super().values()
        values[labels.COLOR] = self._color
        
    def setValue(self, label, value):
        if label == labels.COLOR:
            self.setColor(value)
        else:
            super().setValue(label, value)
            

    
    def create_layout(self):
        
        row = self.layout.rowCount() + 1
        
        self.layout.addWidget(self.materials, row, 0, 1, 2)
        
        
        
        row += 1
        self.layout.addWidget(self.line_width_label, row, 0)
        self.layout.addWidget(self.line_width, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.color_button, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.new_button, row, 0)
        self.layout.addWidget(self.delete_button, row, 1)
        
    def set_callbacks(self):
        super().set_callbacks()                
        
        label = labels.MATERIALS
        callback = lambda _, label=label: self.value_changed(label)        
        self.materials.valueChanged.connect(callback)
        
        
         
        label = labels.LINEWIDTH
        callback = lambda _, label=label: self.value_changed(label)  
        self.line_width.valueChanged.connect(callback)
        
        self.color_button.clicked.connect(self.select_color)
        
        
    def color(self):       
        return self._color
    
    def setColor(self, color):
        self._color = color        
        icon = qta.icon('fa5s.circle', color=color)
        
        self.color_button.setIcon(icon)
        
        self.value_changed(labels.COLOR)
    
    
    def select_color(self):
        color = QColorDialog().getColor()
        if color.isValid():
            self.setColor(color.name())
            
    def setEnabled(self, enabled):
        super().setEnabled(enabled)
        self.new_button.setEnabled(True)
     


        
       

class EditWallsView(ModelItemsWidget):
    explanation = ("To add a Wall select Wall from the toolbar. "
                   "Click and hold the left mouse button to draw a wall")
    
    start_x1, start_y1, start_x2, start_y2 = ['X1 [cm]', 'Y1 [cm]', 
                                              'X2 [cm]', 'Y2 [cm]']
    
    def create_widgets(self):
        super().create_widgets()
        
        self.shielding_label = QLabel("Shielding")
        self.shielding_list = QComboBox()
        
        self.vertices = PositionWidget()
                                
        self.scroll_widget = QScrollBar(QtCore.Qt.Horizontal)
        self.scroll_widget.setPageStep(1)
        
        self.vertex_label = QLabel("Vertex")
        self.vertex_list = QComboBox()
        
        self.index_label = QLabel()
    
    
        
    def value_widgets(self):
        if self._value_widgets is None:
            self._value_widgets = super().value_widgets()
            self._value_widgets[labels.SHIELDING] = self.shielding_list
            #self._value_widgets[labels.VERTICES] = self.vertices
        return self._value_widgets
        
    def set_callbacks(self):
        label = labels.SHIELDING
        callback = lambda _, label=label: self.value_changed(label)
        self.shielding_list.currentTextChanged.connect(callback)
        
        
        # label = labels.VERTICES
        # callback = lambda _, label=label: self.value_changed(label)
        # self.vertices.valueChanged.connect(callback)
                
        
    def setIndex(self, index=None):
        if index is None:
            return
        
        self.scroll_widget.setValue(index)
        self.index_label.setText(f'Wall index {index}')
        
    def create_layout(self): 
        row = 0
        
        self.layout.addWidget(self.scroll_widget, row , 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.index_label, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.shielding_label, row, 0)
        self.layout.addWidget(self.shielding_list, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.vertex_label, row, 0)
        self.layout.addWidget(self.vertex_list, row, 1)
        
        row += 1 
        
        self.layout.addWidget(self.vertices, row, 0, 1, 2)

class EditMaterialsView(ModelItemsWidget):
    LABEL = labels.MATERIALS
    explanation =\
        ('Materials can be changed or added to a limited extend. '
         'Radtracer implements a limited set of materials. Pyshield uses '
         'attenuation and buildup tables for a limited set of '
         'materials. For some materials, like "Concrete-Barite", there is no '
         'buildup table available. For example the buildup table for "Concrete" is '
         'used for "Concrete-Barite" in pyshield. '
         '\n\n'
         'Defining or changing a material is mostly usefull to define a '
         'material that has a (slightly) different density. '
         'Both pyshield and radtracer can handle variations in density accurately.')
        
    
    def create_widgets(self):
        super().create_widgets()
        self.add_list()
        self.add_name()
        self.density_label = QLabel('Density [g/cm^3]')
        self.density_input = QDoubleSpinBox(decimals=3)
        self.attenuation_label = QLabel(labels.ATTENUATION_TABLE)
        self.attenuation_combo = QComboBox()
        self.attenuation_combo.addItems(CONSTANTS.base_materials)
        self.buildup_combo = QComboBox()
        self.buildup_combo.addItems(CONSTANTS.buildup_materials)
        self.buildup_label = QLabel(labels.BUILDUP_TABLE)
        self.radtracer_label = QLabel('Radtracer')
        self.pyshield_label = QLabel('Pyshield')
        self.radtracer_material_label = QLabel(labels.MATERIAL)
        self.radtracer_material_combo = QComboBox()
        self.radtracer_material_combo.addItems(CONSTANTS.base_materials)
        self.new_button = QPushButton('Add', self)
        self.delete_button = QPushButton("Delete")
        
        
        self.attenuation_combo.addItems(CONSTANTS.base_materials)
        self.buildup_combo.addItems(CONSTANTS.buildup_materials)
        
        
    def setValues(self, values):
        super().setValues(values)
        values = values or {}
        if labels.NAME in values.keys()\
            and (values[labels.NAME] == labels.EMPTY_MATERIAL\
                 or values[labels.NAME] in CONSTANTS.base_materials):        
            self.setEnabled(False)
        else:              
            self.setEnabled(True)    
        
    def value_widgets(self):
        if self._value_widgets is None:
            self._value_widgets = super().value_widgets()
            self._value_widgets[labels.DENSITY] = self.density_input
            self._value_widgets[labels.RADTRACER_MATERIAL] = self.radtracer_material_combo
            self._value_widgets[labels.ATTENUATION_TABLE] = self.attenuation_combo
            self._value_widgets[labels.BUILDUP_TABLE] = self.buildup_combo
        return self._value_widgets
            
    def set_callbacks(self):
        super().set_callbacks()
        
        label = labels.DENSITY
        callback = lambda _, label=label: self.value_changed(label)
        self.density_input.valueChanged.connect(callback)
        label = labels.RADTRACER_MATERIAL
        callback = lambda _, label=label: self.value_changed(label)
        self.radtracer_material_combo.currentIndexChanged.connect(callback)
        label = labels.ATTENUATION_TABLE
        callback = lambda _, label=label: self.value_changed(label)
        self.attenuation_combo.currentIndexChanged.connect(callback)
        label = labels.BUILDUP_TABLE
        callback = lambda _, label=label: self.value_changed(label)
        self.buildup_combo.currentIndexChanged.connect(callback)
        
        

    def create_layout(self):
        super().create_layout()
        
        row = self.layout.rowCount() + 1
        
        self.layout.addWidget(self.density_label, row, 0)
        self.layout.addWidget(self.density_input, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.radtracer_label, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.radtracer_material_label, row, 0)
        self.layout.addWidget(self.radtracer_material_combo, row, 1)
        
        
        row += 1
        self.layout.addWidget(self.pyshield_label, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.attenuation_label, row, 0)
        self.layout.addWidget(self.attenuation_combo, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.buildup_label, row, 0)
        self.layout.addWidget(self.buildup_combo, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.new_button, row, 0)
        self.layout.addWidget(self.delete_button, row, 1)
    
    def setEnabled(self, enabled):
        super().setEnabled(enabled)
        self.new_button.setEnabled(True)




class EditPixelSizeView(ModelItemsWidget):  
   
    LABEL = labels.PIXEL_SIZE_CM
    
    explanation =\
("The pixel size can be set by hand or by a measurement on the floor plan " 
 "image. Measurement is done by drawing a line between two points for which "
 "the real world distance in cm is known.")


    def create_widgets(self):
        #self.explanation = QLabel(self.explanation.replace('\n', ' '))
        
        
        self.choose_fixed = QRadioButton("Set Fixed")
        
        self.choose_measured = QRadioButton("Measure")

        
        self.measure_button = QPushButton('Measure On Floorplan')
        
        
        
        self.physical_distance_label = QLabel("Real world distance [cm]")
        self.physical_distance = QDoubleSpinBox(decimals=2)
        self.physical_distance.setRange(-MAX_VALUE, MAX_VALUE)
        
        self.pixel_distance_label = QLabel("Distance [pixels]")
        self.pixel_distance = QLabel("Use Button To Measure")
        
        self.pixel_size_label = QLabel(labels.PIXEL_SIZE_CM)
        self.pixel_size = QDoubleSpinBox(decimals=2)
        self.pixel_size.setRange(-MAX_VALUE, MAX_VALUE)
        
        self.confirm_button = QPushButton("Confirm")
        
        self.choose_fixed.toggled.connect(self.radio_button)
        self.choose_measured.toggled.connect(self.radio_button)
                
        self.vertices = VerticesWidget(self, unit='pixels')
                                      
        
    def set_callbacks(self):
        label = labels.GEOMETRY
        cbk = lambda _, label=label: self.value_changed(label)
        self.choose_fixed.toggled.connect(cbk)
        self.choose_measured.toggled.connect(cbk)
        
        label = labels.REAL_WORLD_DISTANCE_CM
        cbk = lambda _, label=label: self.value_changed(label)
        self.physical_distance.valueChanged.connect(cbk)
        
        label = labels.PIXEL_SIZE_CM
        cbk = lambda _, label=label: self.value_changed(label)
        self.pixel_size.valueChanged.connect(cbk)
        
        label = labels.VERTICES_PIXELS
        cbk = lambda _, label=label: self.value_changed(label)
        self.vertices.valueChanged.connect(cbk)
        
        
    def value_widgets(self):
        if self.choose_fixed.isChecked():
            return {labels.PIXEL_SIZE_CM: self.pixel_size,
                    labels.GEOMETRY: self.choose_fixed}
        
        elif self.choose_measured.isChecked():
            return {labels.REAL_WORLD_DISTANCE_CM: self.physical_distance,
                    labels.VERTICES_PIXELS: self.vertices,
                    labels.GEOMETRY: self.choose_measured,
                    labels.PIXEL_SIZE_CM: self.pixel_size}
    
    def setValue(self, label, value):
        if label == labels.ORIGIN: return
        
        super().setValue(label, value)
        
        if label == labels.VERTICES_PIXELS:            
            vvp = value
            
            d= ((vvp[0][0]-vvp[1][0])**2 + (vvp[0][1] - vvp[1][1])**2)**0.5
            
            safe_set_value_to_widget(self.pixel_distance, round(d,1))
            self.set_pixel_size()
            
        elif label == labels.REAL_WORLD_DISTANCE_CM:
            self.set_pixel_size()
        
            
    def set_pixel_size(self):
        
        try:
            pd = float(safe_get_value_from_widget(self.pixel_distance))
        except ValueError: # self.pixel_distance in initiated as a string
            pd = 0
        
        rwd = safe_get_value_from_widget(self.physical_distance)
        
        if pd > 0 and rwd > 0:
            safe_set_value_to_widget(self.pixel_size, round(rwd/pd, 2))
            
    
    
        
        

    def radio_button(self):        
        if self.choose_measured.isChecked():
            self.set_choose_measured()
        elif self.choose_fixed.isChecked():
            self.set_choose_fixed()
        
    def set_choose_fixed(self):

        self.choose_measured.setChecked(False)
        self.choose_fixed.setChecked(True)
        
        self.measure_button.setEnabled(False)
     
        self.pixel_size_label.setEnabled(True)
        self.pixel_size.setEnabled(True)        
        self.vertices.setEnabled(False)
        
        self.physical_distance_label.setEnabled(False)
        self.physical_distance.setEnabled(False)
        
        self.pixel_distance_label.setEnabled(False)
        self.pixel_distance_label.setEnabled(False)
        
        self.vertices.setEnabled(False)
       
        
    def set_choose_measured(self):

        self.choose_measured.setChecked(True)
        self.choose_fixed.setChecked(False)
        self.measure_button.setEnabled(True)
        self.physical_distance.setEnabled(True)
        self.pixel_size_label.setEnabled(True)
        self.pixel_size.setEnabled(False)
        self.vertices.setEnabled(True)
        self.physical_distance_label.setEnabled(True)
        self.physical_distance.setEnabled(True)
        self.vertices.setEnabled(True)
        
        self.pixel_distance_label.setEnabled(True)
        self.pixel_distance_label.setEnabled(True)
        
        
    def create_layout(self):
        
        row = self.layout.rowCount() + 1
        
        #self.layout.addWidget(self.explanation, row, 0, 1, 2)
        
        #row += 1
        
        self.layout.addWidget(self.choose_fixed, row, 0, 1, 2)

        row += 1
        
        self.layout.addWidget(self.pixel_size_label, row, 0)
        self.layout.addWidget(self.pixel_size, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.choose_measured, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.physical_distance_label, row, 0)
        self.layout.addWidget(self.physical_distance, row, 1)
        
        row += 1
        
        self.layout.addWidget(self.pixel_distance_label, row, 0)
        self.layout.addWidget(self.pixel_distance, row, 1)
        
        row += 1 
        
        self.layout.addWidget(self.vertices, row, 0, 1, 2)
        
        
        row += 1
        
        self.layout.addWidget(self.measure_button, row, 0, 1, 2)
        
        row += 1
        
        self.layout.addWidget(self.confirm_button, row, 0, 1, 2)
        
        row += 1
        
    
        
        self.set_choose_measured()
       
        



       
if __name__ == "__main__":
    from pyrateshield import model_items
    x=Model.load_from_project_file('../../example_projects/SmallProject/project.zip')   
    sequence = x.sources_nm
   
    #sequence.deleteItem(sequence.items[0])
    sequence.items[0].name='test'
    
    x.sources_xray.addItem(model_items.SourceXray())
    x.sources_xray.addItem(model_items.SourceXray())
    app = QApplication([])
    window = QMainWindow()
    widget = EditClearancesView()
    window.setCentralWidget(widget)
    
    #contr = EditShieldingsController(model=x.walls, view=widget)
    #contr = EditShieldingsController(model=x.shieldings, view=widget)
    window.show()    
    app.exec_()


        



import numpy as np
import matplotlib.pyplot as plt
from pyrateshield.constants import CONSTANTS
from pyrateshield.radtracer.radtracer import TransmissionArcher, TransmissionMCNP, MCNP_LOOKUP


if 0:
    for material in CONSTANTS.base_materials:
        if material == 'None': continue
        
        if material == "Lead":
            X = np.linspace(0,0.3,50,endpoint=True)
        else:
            X = np.linspace(0,30,50,endpoint=True)
        
        for params in CONSTANTS.ct:
            transmission = TransmissionArcher(params.archer)      
            plt.plot(X, transmission.get(material, X), '-', label="CT "+str(params.kvp))

        for params in CONSTANTS.xray:
            transmission = TransmissionArcher(params.archer)      
            plt.plot(X, transmission.get(material, X), '--', label="xray "+str(params.kvp))
        
        plt.title(material)
        plt.xlabel("Thickness [cm]")
        plt.ylabel("Transmission")
        plt.yscale("log")
        plt.ylim((1E-5,1))
        plt.legend(title="kVp")
        plt.show()

def arch(a,b,g,x):
    return ( (1 + b/a)*np.exp(a*g*x) - b/a )**(-1/g) 

cycle = plt.rcParams['axes.prop_cycle'].by_key()['color']
for material in CONSTANTS.base_materials:
    if material == 'None': continue
    
    if material == "Lead":
        X = np.arange(0,18.1,0.1)
        xlim = (-0.1, 5)
        ylim = (1E-5, 2)
    
        #a, b, g = (1.322, 148.5, 1.488)
        #plt.plot(X, arch(a,b,g,X), ":", color="k", label="I-123 Archer", lw=3)
    else:
        X = np.arange(0,181,1)
        xlim = (-0.5, 15)
        ylim = (1E-2, 1.1)
        
        #a, b, g = (0.1866, 1253, 74.93)
        #plt.plot(X, arch(a,b,g,X), ":", color="k", label="I-123 Archer", lw=3)
        
    for i, isotope in enumerate(CONSTANTS.isotopes):
        if isotope.name == "Ge-68": #Ge-68 is just a copy of Ga-68
            continue
        
        lw = 2 if i//10 else 1.2
        
        for self_shielding in ["None", "Body"]:
            transmission = TransmissionMCNP(MCNP_LOOKUP[isotope.name][self_shielding])            
            ls = "-" if self_shielding == "None" else "--"
            lbl = isotope.name if self_shielding == "None" else None
            plt.plot(X, transmission.get(material, X), ls, color=cycle[i%10], label=lbl, lw=lw)
    
    
    
    
    
    plt.title(material + " - dashed: patient self-shielding")
    plt.xlabel("Thickness [cm]")
    plt.ylabel("Transmission")
    plt.yscale("log")
    plt.xlim(xlim)
    plt.ylim(ylim)
    plt.legend()
    plt.show()
